# mypy: ignore-errors
from typing import Any

import torch
import torch.utils._pytree as pytree
from torch._subclasses.fake_tensor import is_fake
from torch.testing._internal.two_tensor import TwoTensor
from torch.utils._python_dispatch import return_and_correct_aliasing


class WrapperSubclass(torch.Tensor):
    @staticmethod
    def __new__(cls, a, outer_size=None, outer_stride=None):
        if outer_size is None:
            outer_size = a.size()
        if outer_stride is None:
            outer_stride = a.stride()

        kwargs = {}
        kwargs["strides"] = outer_stride
        kwargs["storage_offset"] = a.storage_offset()
        kwargs["device"] = a.device
        kwargs["layout"] = a.layout
        kwargs["requires_grad"] = a.requires_grad
        kwargs["dtype"] = a.dtype
        out = torch.Tensor._make_wrapper_subclass(cls, outer_size, **kwargs)

        return out

    def __init__(self, a, outer_size=None, outer_stride=None):
        self.a = a

    def __repr__(self):
        return f"WrapperSubclass({repr(self.a)})"

    def __tensor_flatten__(self):
        return ["a"], None

    @staticmethod
    def __tensor_unflatten__(inner_tensors, meta, outer_size, outer_stride):
        if meta is not None:
            raise AssertionError("Expected meta to be None")
        a = inner_tensors["a"]
        if is_fake(a):
            if outer_size is None:
                raise AssertionError("Expected outer_size to not be None")
            if outer_stride is None:
                raise AssertionError("Expected outer_stride to not be None")
        return WrapperSubclass(a, outer_size, outer_stride)

    @classmethod
    def __torch_dispatch__(cls, func, types, args, kwargs):
        if kwargs is None:
            kwargs = {}
        args_a = pytree.tree_map_only(WrapperSubclass, lambda x: x.a, args)

        kwargs_a = pytree.tree_map_only(WrapperSubclass, lambda x: x.a, kwargs)

        out_a = func(*args_a, **kwargs_a)
        out_a_flat, spec = pytree.tree_flatten(out_a)
        out_flat = [
            WrapperSubclass(o_a) if isinstance(o_a, torch.Tensor) else o_a
            for o_a in out_a_flat
        ]
        out = pytree.tree_unflatten(out_flat, spec)
        from torch._higher_order_ops.cond import cond_op

        if func is cond_op:
            return out
        else:
            return return_and_correct_aliasing(func, args, kwargs, out)

    def __coerce_same_metadata_as_tangent__(
        self, expected_metadata: Any, expected_type: type | None = None
    ):
        if expected_type is type(self.a):
            return self.a
        elif expected_type is TwoTensor:
            return TwoTensor(self.a, self.a.clone())

        return None



# Owner(s): ["module: inductor"]
import atexit
import contextlib
import functools
import itertools
import math
import os
import sys
import unittest
from collections import defaultdict
from enum import Enum
from functools import partial
from unittest.mock import patch

import torch
from torch._dispatch.python import enable_python_dispatcher
from torch._inductor.test_case import run_tests, TestCase
from torch._subclasses.fake_tensor import (
    DataDependentOutputException,
    DynamicOutputShapeException,
    FakeTensorMode,
)
from torch.testing._internal.common_cuda import SM80OrLater
from torch.testing._internal.common_device_type import (
    instantiate_device_type_tests,
    onlyNativeDeviceTypes,
    OpDTypes,
    ops,
    skipCPUIf,
    skipXPUIf,
)
from torch.testing._internal.common_methods_invocations import op_db, skipOps
from torch.testing._internal.common_utils import (
    IS_CI,
    IS_LINUX,
    IS_MACOS,
    IS_WINDOWS,
    IS_X86,
    skipCUDAMemoryLeakCheckIf,
    skipIfCrossRef,
    skipIfTorchDynamo,
    suppress_warnings,
    TEST_MKL,
    TEST_WITH_ASAN,
    TEST_WITH_ROCM,
)
from torch.testing._internal.inductor_utils import (
    GPU_TYPE,
    HAS_CPU,
    has_triton,
    HAS_XPU_AND_TRITON,
    maybe_skip_size_asserts,
)
from torch.testing._internal.triton_utils import requires_gpu_and_triton
from torch.utils._dtype_abbrs import dtype_abbrs
from torch.utils._python_dispatch import TorchDispatchMode
from torch.utils._pytree import tree_map


try:
    try:
        from .test_torchinductor import check_model, check_model_gpu
    except ImportError:
        from test_torchinductor import (  # @manual=fbcode//caffe2/test/inductor:test_inductor-library
            check_model,
            check_model_gpu,
        )
except (unittest.SkipTest, ImportError) as e:
    sys.stderr.write(f"{type(e)}: {e}\n")
    if __name__ == "__main__":
        sys.exit(0)
    raise

if IS_WINDOWS and IS_CI:
    # TODO(xuhancn) : improve the compiler build performance on windows.
    sys.stderr.write(
        "This UT is too slow on windows, and will cause out of time in CI. So skip it now.\n"
    )
    if __name__ == "__main__":
        sys.exit(0)
    raise unittest.SkipTest("skip slow test")

bf16 = torch.bfloat16  # not tested
f64 = torch.float64
f32 = torch.float32
f16 = torch.float16
i8 = torch.int8  # not tested
i16 = torch.int16  # not tested
i32 = torch.int32
i64 = torch.int64
b8 = torch.bool
u8 = torch.uint8  # not tested except upsampling and interpolate ops
u16 = torch.uint16  # not tested
u32 = torch.uint32  # not tested
u64 = torch.uint64  # not tested

_ops = partial(
    ops,
    dtypes=OpDTypes.supported,
    allowed_dtypes=[f16, f32, f64, i32, i64, b8, u8, u16, u32, u64],
)

# Success forces pass; failure forces fail; skip unconditionally skips testing
ExpectedTestResult = Enum("ExpectedTestResult", ("SUCCESS", "XFAILURE", "SKIP"))

COLLECT_EXPECT = os.getenv("PYTORCH_COLLECT_EXPECT", "0") == "1"
ALL_SAMPLES = os.getenv("PYTORCH_ALL_SAMPLES", "0") == "1"
START = os.getenv("PYTORCH_TEST_RANGE_START", None)
END = os.getenv("PYTORCH_TEST_RANGE_END", None)

if START is not None or END is not None:
    if END is None:
        raise AssertionError("END must be set when START is set")
    if START is None:
        raise AssertionError("START must be set when END is set")
    START = int(START)
    END = int(END)
    if START >= END:
        raise AssertionError(f"START ({START}) must be less than END ({END})")
else:
    START = 0
    END = len(op_db)

seen_failed = defaultdict(set)
failed_reasons = defaultdict(set)


def print_seen():
    expected_failures = defaultdict(list)

    def fmt_dtypes(dtypes):
        r = ", ".join(sorted(dtype_abbrs[d] for d in dtypes))
        return "{" + r + "}"

    def sort_key(kv):
        k, _ = kv
        _, op = k
        if isinstance(op, tuple):
            return op
        else:
            return op, ""

    for (device_type, op), failed_dtypes in sorted(seen_failed.items(), key=sort_key):
        key = device_type, op
        reasons = ""
        if failed_reasons[key]:

            def maybe_truncate(x, length=80):
                x = str(x).replace("\n", " ")

                idx = x.find("\\n")
                if idx >= 0:
                    x = f"{x[:idx]}..."
                if len(x) > length:
                    return f"{x[: length - 3]}..."
                return x

            reasons = sorted(set(map(maybe_truncate, failed_reasons[key])))
            reasons = "  # " + ", ".join(reasons)

        if failed_dtypes:

            def format_op(op):
                if isinstance(op, tuple):
                    return f'("{op[0]}", "{op[1]}")'
                else:
                    return f'"{op}"'

            expected_failures[device_type].append(
                f"    {format_op(op)}: {fmt_dtypes(failed_dtypes)},{reasons}"
            )

    for device_type in ("cpu", GPU_TYPE):
        expected_failures[device_type]
        nl = "\n"
        print(
            f"""
inductor_expected_failures_single_sample[\"{device_type}\"] = {{
{nl.join(expected_failures[device_type])}
}}
"""
        )


if COLLECT_EXPECT:
    atexit.register(print_seen)

# Note, in these skip/xfail dictionaries use a string as the key
# for the default test, and a tuple of two strings for variants

inductor_skips = defaultdict(dict)


inductor_skips["cpu"] = {
    "linalg.ldl_factor": {f32, f64},  # flaky
    "nn.functional.cosine_embedding_loss": {b8},  # flaky
    ("index_reduce", "prod"): {f16},  # flaky
    ("index_reduce", "mean"): {f16},  # flaky
}

if IS_MACOS and IS_X86:
    inductor_skips["cpu"]["rsqrt"] = {b8, i32}
    inductor_skips["cpu"]["nn.functional.multi_margin_loss"] = {
        b8,
        f16,
        f32,
        f64,
        i32,
        i64,
    }

inductor_skips["cuda"] = {
    # Jiterator kernel is not expected to work with inductor
    "jiterator_2inputs_2outputs": {b8, f16, f32, f64, i32, i64},
    "jiterator_4inputs_with_extra_args": {b8, f16, f32, f64, i32, i64},
    "jiterator_binary": {b8, f16, f32, f64, i32, i64},
    "jiterator_binary_return_by_ref": {b8, f16, f32, f64, i32, i64},
    "jiterator_unary": {b8, f16, f32, f64, i32, i64},
    # flaky
    "nn.functional.cosine_embedding_loss": {b8},
    "native_batch_norm": {f16, f32, f64},
    "_native_batch_norm_legit": {f16, f32, f64},
    "_batch_norm_with_update": {f16, f32, f64},
}

if not SM80OrLater:
    inductor_skips["cuda"]["bfloat16"] = {b8, f16, f32, f64, i32, i64}

if TEST_WITH_ROCM:
    # Tensors are not alike
    inductor_skips["cuda"]["logcumsumexp"] = {f32}
    inductor_skips["cuda"]["special.modified_bessel_i1"] = {f64}

inductor_skips["xpu"] = {}

# torch-xpu-ops: #2956
inductor_skips["xpu"]["lu"] = {f32}
inductor_skips["xpu"]["nn.functional.linear"] = {f16}
inductor_skips["xpu"]["masked.cumprod"] = {f16}

inductor_expected_failures_single_sample = defaultdict(dict)

inductor_expected_failures_single_sample["cpu"] = {
    "_upsample_bilinear2d_aa": {f32, f64},
    "cholesky": {f32, f64},
    "complex": {f16},
    "resize_": {b8, f16, f32, f64, i32, i64},
    "resize_as_": {b8, f16, f32, f64, i32, i64},
    "histc": {f16},
    "multinomial": {f16, f32, f64},
    "nonzero_static": {b8, f16, f32, f64, i32, i64},
    ("normal", "in_place"): {f16, f32, f64},
    ("normal", "number_mean"): {f16, f32, f64},
    "normal": {f16, f32, f64},
    ("sparse.mm", "reduce"): {f32, f64, f16},
    "sparse.sampled_addmm": {f32, f64},
    "to_sparse": {
        f32,
        f64,
    },  # NYI: could not find kernel for aten.view.default at dispatch key DispatchKey.SparseCPU
    "view_as_complex": {f16},
}


inductor_expected_failures_single_sample["cuda"] = {
    "_upsample_bilinear2d_aa": {f16, f32, f64},
    "cholesky": {f32, f64},
    "multinomial": {f16, f32, f64},
    ("normal", "in_place"): {f16, f32, f64},
    ("normal", "number_mean"): {f16, f32, f64},
    "normal": {f16, f32, f64},
    "sparse.sampled_addmm": {f32, f64},
    "torch.ops.aten._flash_attention_forward": {f16},
    "torch.ops.aten._efficient_attention_forward": {f16, f32},
    "to_sparse": {
        b8,
        f16,
        f32,
        f64,
        i32,
        i64,
    },  # NYI: could not find kernel for aten.view.default at dispatch key DispatchKey.SparseCUDA
}

inductor_expected_failures_single_sample["xpu"] = {
    "_upsample_bilinear2d_aa": {f16, f32, f64},
    "cholesky": {f32, f64},
    "multinomial": {f16, f32, f64},
    ("normal", "in_place"): {f16, f32, f64},
    ("normal", "number_mean"): {f16, f32, f64},
    "normal": {f16, f32, f64},
    "sparse.sampled_addmm": {f32, f64},
    "tan": {f16},
    "torch.ops.aten._flash_attention_forward": {f16},
    "torch.ops.aten._efficient_attention_forward": {f16, f32},
    "to_sparse": {
        b8,
        f16,
        f32,
        f64,
        i32,
        i64,
    },  # align with cuda.
}


# intentionally not handled
intentionally_not_handled = {
    "resize_": {b8, f16, f32, f64, i32, i64},
    "resize_as_": {b8, f16, f32, f64, i32, i64},
}
# This is only fixed when this config is set
# We should eventually always turn it on
import torch._functorch.config as functorch_config


if not functorch_config.view_replay_for_aliased_outputs:
    intentionally_not_handled['("as_strided", "partial_views")'] = {
        b8,
        f16,
        f32,
        f64,
        i32,
        i64,
    }

inductor_expected_failures_single_sample["cuda"].update(intentionally_not_handled)
inductor_expected_failures_single_sample["xpu"].update(intentionally_not_handled)


inductor_gradient_expected_failures_single_sample = defaultdict(dict)

inductor_gradient_expected_failures_single_sample["cuda"] = {}
inductor_gradient_expected_failures_single_sample["xpu"] = {}

if not TEST_MKL:
    inductor_expected_failures_single_sample["cpu"].update({})

inductor_should_fail_with_exception = defaultdict(dict)
inductor_should_fail_with_exception["cpu"] = {}
inductor_should_fail_with_exception["cuda"] = {}
inductor_should_fail_with_exception["xpu"] = {}


def get_skips_and_xfails(from_dict, xfails=True):
    retval = set()
    for device, d in from_dict.items():
        for op, dtypes in d.items():
            if type(op) is tuple:
                op, variant_name = op
            else:
                variant_name = ""
            retval.add((op, variant_name, device, tuple(dtypes), xfails))
    return retval


# Note: if you get a "AssertionError: Couldn't find OpInfo for ..." error for an OpInfo you are sure
# exists, you might be trying to use a test variant and you need to replace, for example,
# "max.reduction_no_dim" with ("max", "reduction_no_dim") as the key of one of these dictionaries
test_skips_or_fails = (
    get_skips_and_xfails(inductor_skips, xfails=False)
    | get_skips_and_xfails(inductor_expected_failures_single_sample, xfails=True)
    | get_skips_and_xfails(
        inductor_gradient_expected_failures_single_sample, xfails=True
    )
)


def wrapper_noop_set_seed(op, *args, **kwargs):
    return op(*args, **kwargs)


wrapper_noop_set_seed_decorator = patch(
    "torch.testing._internal.common_methods_invocations.wrapper_set_seed",
    wrapper_noop_set_seed,
)

# key can be either op_name, or (op_name, dtype)
inductor_override_kwargs = defaultdict(dict)

inductor_override_kwargs["cpu"] = {
    # the return value of empty is undefined
    "empty": {"assert_equal": False},
    "empty_permuted": {"assert_equal": False},
    "empty_like": {"assert_equal": False},
    "new_empty": {"assert_equal": False},
    "empty_strided": {"assert_equal": False},
    "new_empty_strided": {"assert_equal": False},
    "randn": {"assert_equal": False},
    ("nn.functional.multilabel_soft_margin_loss", f16): {
        "atol": 3e-4,
        "rtol": 0.002,
    },
    ("nn.functional.triplet_margin_loss", f16): {"atol": 3e-4, "rtol": 0.003},
    ("nn.functional.triplet_margin_with_distance_loss", f16): {
        "atol": 3e-4,
        "rtol": 0.003,
    },
    ("softmax", f16): {"atol": 1e-4, "rtol": 0.02},
    ("polygamma.polygamma_n_0", f32): {"atol": 1e-3, "rtol": 1e-4},
    ("polygamma.polygamma_n_1", f32): {"atol": 1e-3, "rtol": 1e-4},
    ("polygamma.polygamma_n_2", f32): {"atol": 1e-3, "rtol": 1e-4},
    ("polygamma.polygamma_n_3", f32): {"atol": 1e-3, "rtol": 1e-4},
    ("polygamma.polygamma_n_4", f32): {"atol": 1e-3, "rtol": 1e-4},
    ("special.polygamma.special_polygamma_n_0", f32): {
        "atol": 1e-3,
        "rtol": 1e-4,
    },
    ("_unsafe_masked_index_put_accumulate", f16): {"atol": 1e-4, "rtol": 0.01},
    # Following tests are failing with strict comparison but atol=1 is acceptable due roundings errors
    ("nn.functional.interpolate.bilinear", u8): {"atol": 1, "rtol": 0},
    ("nn.functional.upsample_bilinear", u8): {"atol": 1, "rtol": 0},
    ("nn.functional.interpolate.bicubic", u8): {"atol": 1, "rtol": 0},
    # High atol due to precision loss
    ("nn.functional.interpolate.bicubic", f32): {"atol": 5e-3, "rtol": 0},
}

inductor_override_kwargs["cuda"] = {
    # the return value of empty is undefined
    "empty": {"assert_equal": False},
    "empty_permuted": {"assert_equal": False},
    "empty_like": {"assert_equal": False},
    "new_empty": {"assert_equal": False},
    "empty_strided": {"assert_equal": False},
    "new_empty_strided": {"assert_equal": False},
    "randn": {"assert_equal": False},
    ("cross", f16): {"reference_in_float": True},
    ("linalg.cross", f16): {"reference_in_float": True},
    ("addr", f16): {"reference_in_float": True},
    ("baddbmm", f16): {"atol": 2e-3, "rtol": 0.002},  # decomp affects accuracy
    ("angle", f64): {"reference_in_float": True},
    ("asin", f16): {"reference_in_float": True},
    ("atanh", f16): {"reference_in_float": True},
    "cauchy": {"reference_in_float": True},
    ("cummax", f16): {"atol": 5e-4, "rtol": 0.002},
    ("cumsum", f16): {"reference_in_float": True},
    "cumprod": {"reference_in_float": True, "atol": 7e-5, "rtol": 0.002},
    "logcumsumexp": {"grad_atol": 8e-4, "grad_rtol": 0.001},
    ("logcumsumexp", f16): {"grad_atol": 3e-3, "grad_rtol": 0.01},
    "exponential": {"reference_in_float": True},
    "geometric": {"reference_in_float": True},
    ("kron", f16): {"reference_in_float": True},
    "log_normal": {"reference_in_float": True},
    ("masked.softmin", f16): {"atol": 1e-4, "rtol": 0.01},
    ("nn.functional.adaptive_avg_pool2d", f16): {
        "reference_in_float": True,
        "atol": 2e-5,
        "rtol": 0.02,
    },
    ("nn.functional.batch_norm", f16): {"reference_in_float": True},
    ("nn.functional.batch_norm.without_cudnn", f16): {"reference_in_float": True},
    ("nn.functional.cosine_similarity", f16): {"reference_in_float": True},
    ("nn.functional.instance_norm", f16): {"reference_in_float": True},
    ("nn.functional.linear", f16): {"atol": 3e-4, "rtol": 0.01},
    ("nn.functional.local_response_norm", f16): {"reference_in_float": True},
    ("nn.functional.normalize", f16): {"atol": 1e-3, "rtol": 0.05},
    ("nn.functional.rms_norm", f16): {"reference_in_float": True},
    ("nn.functional.soft_margin_loss", f16): {"reference_in_float": True},
    ("nn.functional.softmin", f16): {"atol": 1e-4, "rtol": 0.01},
    ("nn.functional.softsign", f16): {"reference_in_float": True},
    ("nn.functional.tanhshrink", f16): {"atol": 3e-4, "rtol": 0.001},
    ("outer", f16): {"reference_in_float": True},
    ("round.decimals_3", f16): {"reference_in_float": True},
    ("nn.functional.triplet_margin_loss", f16): {"atol": 1e-4, "rtol": 0.02},
    ("nn.functional.triplet_margin_with_distance_loss", f16): {
        "atol": 1e-4,
        "rtol": 0.02,
    },
    ("sinc", f16): {"atol": 0.008, "rtol": 0.002},
    ("torch.ops.aten._safe_softmax.default", f16): {"atol": 5e-4, "rtol": 0.02},
    ("softmax", f16): {"atol": 1e-4, "rtol": 0.02},
    ("_softmax_backward_data", f16): {"atol": 0.008, "rtol": 0.002},
    ("special.log_ndtr", f64): {"atol": 1e-6, "rtol": 1e-5},
    ("std_mean.unbiased", f16): {"reference_in_float": True},
    "uniform": {"reference_in_float": True},
    ("_unsafe_masked_index_put_accumulate", f16): {"atol": 1e-4, "rtol": 0.01},
    # High atol due to precision loss
    ("nn.functional.interpolate.bilinear", f64): {"atol": 5e-4, "rtol": 0},
    ("nn.functional.upsample_bilinear", f64): {"atol": 5e-4, "rtol": 0},
    ("nn.functional.interpolate.bicubic", f64): {"atol": 1e-3, "rtol": 0},
    # Unreasonably high atol requirement:
    ("index_reduce.mean", f16): {"check_gradient": False},
    ("index_reduce.mean", f32): {"check_gradient": False},
    ("index_reduce.mean", f64): {"check_gradient": False},
    # Gradient contains non-finite entries:
    ("index_reduce.amin", f64): {"check_gradient": False},
    ("index_reduce.amin", f32): {"check_gradient": False},
    ("index_reduce.amin", f16): {"check_gradient": False},
    ("index_reduce.amax", f64): {"check_gradient": False},
    ("index_reduce.amax", f32): {"check_gradient": False},
    ("index_reduce.amax", f16): {"check_gradient": False},
    ("tanh", f16): {"atol": 1e-4, "rtol": 1e-2},
    ("_unsafe_masked_index", f16): {
        "reference_in_float": True,
        "atol": 3e-4,
        "rtol": 2e-3,
    },
    ("nn.functional.interpolate.linear", f16): {"reference_in_float": True},
    ("nn.functional.prelu", f16): {
        "reference_in_float": True,
        "atol": 1e-3,
        "rtol": 4e-3,
    },
    ("addmm", f16): {"reference_in_float": True},
    ("logaddexp", f16): {"reference_in_float": True},
    ("std_mean", f16): {"reference_in_float": True},
    ("hypot", f16): {"reference_in_float": True, "atol": 3e-4, "rtol": 2e-3},
    ("cummin", f16): {"reference_in_float": True, "atol": 5e-5, "rtol": 2e-3},
    ("unfold_copy", f16): {"reference_in_float": True, "atol": 2e-5, "rtol": 1e-2},
    ("nn.functional.upsample_bilinear", f16): {
        "reference_in_float": True,
        "atol": 1e-4,
        "rtol": 2e-3,
    },
    ("nn.functional.embedding_bag", f16): {
        "reference_in_float": True,
        "atol": 1e-4,
        "rtol": 1e-2,
    },
    ("fft.irfft2", f16): {
        "reference_in_float": True,
        "atol": 1e-4,
        "rtol": 7e-1,
    },
    ("fft.irfftn", f16): {
        "reference_in_float": True,
        "atol": 1e-4,
        "rtol": 7e-1,
    },
}

inductor_override_kwargs["xpu"] = {
    # the return value of empty is undefined
    "empty": {"assert_equal": False},
    "empty_permuted": {"assert_equal": False},
    "empty_like": {"assert_equal": False},
    "new_empty": {"assert_equal": False},
    "empty_strided": {"assert_equal": False},
    "new_empty_strided": {"assert_equal": False},
    "randn": {"assert_equal": False},
    # XPU
    ("cross", f16): {"reference_in_float": True},
    ("addr", f16): {"reference_in_float": True},
    ("baddbmm", f16): {"atol": 2e-3, "rtol": 0.002},  # decomp affects accuracy
    ("angle", f64): {"reference_in_float": True},
    ("asin", f16): {"reference_in_float": True},
    ("asin", f32): {"reference_in_float": True, "atol": 1e-4, "rtol": 1e-4},
    ("atanh", f16): {"reference_in_float": True},
    "cauchy": {"reference_in_float": True},
    ("cummax", f16): {"atol": 5e-4, "rtol": 0.002},
    ("cumsum", f16): {"reference_in_float": True},
    "cumprod": {"reference_in_float": True, "atol": 7e-5, "rtol": 0.002},
    ("dot", f16): {"atol": 1e-5, "rtol": 0.002},
    "logcumsumexp": {
        "atol": 5e-5,
        "rtol": 0.005,
        "grad_atol": 8e-4,
        "grad_rtol": 0.001,
    },
    ("logcumsumexp", f16): {"grad_atol": 4e-3, "grad_rtol": 0.01},
    "exponential": {"reference_in_float": True},
    "geometric": {"reference_in_float": True},
    ("kron", f16): {"reference_in_float": True},
    ("linalg.cross", f16): {"reference_in_float": True},
    ("linalg.vecdot", f16): {"atol": 1e-5, "rtol": 2e-2},
    "log_normal": {"reference_in_float": True},
    ("logsumexp", f16): {"atol": 1e-5, "rtol": 1e-2},
    ("masked.cumprod", f16): {"reference_in_float": True, "atol": 1e-5, "rtol": 5e-2},
    ("masked.cumsum", f16): {"atol": 1e-5, "rtol": 5e-3},
    ("masked.softmin", f16): {"atol": 1e-4, "rtol": 0.01},
    ("masked.softmax", f16): {"atol": 2e-4, "rtol": 0.01},
    ("masked.var", f16): {"atol": 2e-5, "rtol": 5e-3},
    ("native_batch_norm", f64): {"atol": 1e-7, "rtol": 1e-5},
    ("_native_batch_norm_legit", f64): {"atol": 1e-7, "rtol": 5e-6},
    ("_batch_norm_with_update", f64): {"atol": 1e-7, "rtol": 1e-6},
    ("native_layer_norm", f16): {"atol": 5e-3, "rtol": 5e-3},
    ("native_layer_norm", f32): {"atol": 5e-3, "rtol": 5e-3},
    ("nn.functional.batch_norm", f16): {"reference_in_float": True},
    ("nn.functional.batch_norm", f64): {"atol": 1e-6, "rtol": 1e-6},
    ("nn.functional.batch_norm.without_cudnn", f16): {"reference_in_float": True},
    ("nn.functional.conv1d", f16): {"atol": 1e-5, "rtol": 6e-3},
    ("nn.functional.conv3d", f16): {"atol": 1e-5, "rtol": 2e-3},
    ("nn.functional.conv_transpose2d", f16): {"atol": 1e-5, "rtol": 2e-3},
    ("nn.functional.conv_transpose3d", f16): {"atol": 1e-5, "rtol": 5e-3},
    ("nn.functional.cosine_embedding_loss", f16): {"atol": 1e-5, "rtol": 2e-3},
    ("nn.functional.cosine_similarity", f16): {
        "reference_in_float": True,
        "atol": 1e-5,
        "rtol": 5e-3,
    },
    ("nn.functional.instance_norm", f16): {"reference_in_float": True},
    ("nn.functional.instance_norm", f64): {"atol": 1e-6, "rtol": 1e-6},
    ("nn.functional.layer_norm", f16): {"atol": 5e-3, "rtol": 2e-3},
    ("nn.functional.layer_norm", f32): {"atol": 5e-5, "rtol": 2e-3},
    ("nn.functional.local_response_norm", f16): {"reference_in_float": True},
    ("nn.functional.multilabel_soft_margin_loss", f16): {
        "atol": 3e-4,
        "rtol": 2e-3,
    },
    ("nn.functional.normalize", f16): {"atol": 1e-3, "rtol": 0.05},
    ("nn.functional.rms_norm", f16): {"reference_in_float": True},
    ("nn.functional.soft_margin_loss", f16): {"reference_in_float": True},
    ("nn.functional.softmin", f16): {"atol": 1e-4, "rtol": 0.01},
    ("nn.functional.softsign", f16): {
        "reference_in_float": True,
        "atol": 1e-5,
        "rtol": 0.005,
    },
    ("nn.functional.tanhshrink", f16): {"atol": 3e-4, "rtol": 0.001},
    ("outer", f16): {"reference_in_float": True},
    ("round.decimals_3", f16): {"reference_in_float": True},
    ("nn.functional.triplet_margin_loss", f16): {"atol": 1e-4, "rtol": 0.02},
    ("nn.functional.triplet_margin_with_distance_loss", f16): {
        "atol": 1e-4,
        "rtol": 0.02,
    },
    ("remainder", f16): {"atol": 1e-4, "rtol": 0.005},
    ("sinc", f16): {"atol": 0.008, "rtol": 0.002},
    ("softmax", f16): {"atol": 1e-4, "rtol": 0.02},
    ("_softmax_backward_data", f16): {"atol": 0.008, "rtol": 0.002},
    ("special.log_ndtr", f64): {"atol": 1e-6, "rtol": 1e-5},
    ("std_mean.unbiased", f16): {
        "reference_in_float": True,
        "atol": 5e-5,
        "rtol": 5e-3,
    },
    ("trapezoid", f16): {"atol": 1e-5, "rtol": 5e-3},
    ("trapz", f16): {"atol": 1e-5, "rtol": 5e-3},
    "uniform": {"reference_in_float": True},
    ("var_mean", f16): {"atol": 1e-5, "rtol": 2e-3},
    ("var_mean.unbiased", f16): {"atol": 1e-5, "rtol": 2e-3},
    ("vdot", f16): {"atol": 1e-5, "rtol": 2e-3},
    # Following tests are failing with strict comparison but atol=1 is acceptable due roundings errors
    # High atol due to precision loss
    ("nn.functional.interpolate.bilinear", f64): {"atol": 5e-4, "rtol": 0},
    ("nn.functional.upsample_bilinear", f64): {"atol": 5e-4, "rtol": 0},
    ("nn.functional.interpolate.bicubic", f64): {"atol": 1e-3, "rtol": 0},
    # Unreasonably high atol requirement:
    ("index_reduce.mean", f16): {"check_gradient": False},
    ("index_reduce.mean", f32): {"check_gradient": False},
    ("index_reduce.mean", f64): {"check_gradient": False},
    # Gradient contains non-finite entries:
    ("index_reduce.amin", f64): {"check_gradient": False},
    ("index_reduce.amin", f32): {"check_gradient": False},
    ("index_reduce.amin", f16): {"check_gradient": False},
    ("index_reduce.amax", f64): {"check_gradient": False},
    ("index_reduce.amax", f32): {"check_gradient": False},
    ("index_reduce.amax", f16): {"check_gradient": False},
    ("tanh", f16): {"atol": 1e-4, "rtol": 1e-2},
    ("nn.functional.embedding_bag", f32): {"check_gradient": False},
    ("nn.functional.embedding_bag", f64): {"check_gradient": False},
    ("_unsafe_masked_index_put_accumulate", f16): {"atol": 1e-4, "rtol": 0.01},
    ("_unsafe_masked_index", f16): {
        "reference_in_float": True,
        "atol": 3e-4,
        "rtol": 2e-3,
    },
    ("nn.functional.interpolate.linear", f16): {"reference_in_float": True},
    ("nn.functional.prelu", f16): {
        "reference_in_float": True,
        "atol": 1e-3,
        "rtol": 4e-3,
    },
    ("addmm", f16): {"reference_in_float": True},
    ("logaddexp", f16): {"reference_in_float": True},
    ("std_mean", f16): {"reference_in_float": True},
    ("hypot", f16): {"reference_in_float": True, "atol": 3e-4, "rtol": 2e-3},
    ("cummin", f16): {"reference_in_float": True, "atol": 5e-5, "rtol": 2e-3},
    ("unfold_copy", f16): {"reference_in_float": True, "atol": 2e-5, "rtol": 1e-2},
    ("nn.functional.upsample_bilinear", f16): {
        "reference_in_float": True,
        "atol": 1e-4,
        "rtol": 2e-3,
    },
    ("nn.functional.embedding_bag", f16): {
        "check_gradient": False,
        "atol": 1e-4,
        "rtol": 1e-2,
    },
    ("nn.functional.max_pool2d", f16): {
        "reference_in_float": True,
        "atol": 1e-4,
        "rtol": 2e-3,
    },
    ("nn.functional.unfold", f16): {
        "reference_in_float": True,
    },
    # Reference crash on Intel LTS2 driver.
    ("nn.functional.interpolate.trilinear", f32): {
        "check_gradient": False,
    },
    # Reference crash on Intel LTS2 driver.
    ("nn.functional.interpolate.trilinear", f64): {
        "check_gradient": False,
    },
}
if TEST_WITH_ROCM:
    inductor_override_kwargs["cuda"].update(
        {("cummin", f16): {"atol": 1e-3, "rtol": 1e-5}}
    )


# Test with one sample only for following ops
inductor_one_sample = defaultdict(dict)

inductor_one_sample["cpu"] = {
    "_segment_reduce.lengths": {f16},
    "_segment_reduce.offsets": {f16},
    "addmv": {f16},
    "as_strided.partial_views": {f16},
    "corrcoef": {f16},
    "diff": {f16},
    "einsum": {f16, i32},
    "gradient": {f16},
    "histogram": {f32, f64},
    "histogramdd": {f32, f64},
    "index_put": {f16, f32, f64},
    "linalg.eig": {f32, f64},
    "linspace": {f16, i32, i64},
    "linspace.tensor_overload": {f16, f32, f64, i32, i64},
    "logspace": {f16},
    "logspace.tensor_overload": {f16, f32, f64, i32, i64},
    "masked_logsumexp": {i64},
    "max_pool2d_with_indices_backward": {f16, f32, f64},
    "new_empty_strided": {f16},
    "nn.functional.adaptive_avg_pool3d": {f16},
    "nn.functional.adaptive_max_pool1d": {f16, f32},
    "nn.functional.adaptive_max_pool2d": {f16, f32},
    "nn.functional.bilinear": {f16},
    "nn.functional.conv_transpose1d": {f16},
    "nn.functional.conv_transpose2d": {f16},
    "nn.functional.conv_transpose3d": {f16},
    "nn.functional.cosine_similarity": {f16},
    "nn.functional.cross_entropy": {f16, f32, f64},
    "nn.functional.gaussian_nll_loss": {f16},
    "nn.functional.grid_sample": {f32, f64, f16},
    "nn.functional.interpolate.area": {f16},
    "nn.functional.nll_loss": {f16, f32, f64},
    "normal": {f16, f32, f64},
    "put": {f16, f32, f64},
    "take": {b8, f16, f32, f64, i32, i64},
}

inductor_one_sample["cuda"] = {
    "_segment_reduce.lengths": {f16},
    "_segment_reduce.offsets": {f16},
    "addmv": {f16},
    "as_strided.partial_views": {f16},
    "corrcoef": {f16},
    "diff": {f16},
    "einsum": {f16, i32},
    "gradient": {f16},
    "histogram": {f32, f64},
    "histogramdd": {f32, f64},
    "index_put": {f16, f32, f64},
    "linalg.eig": {f32, f64},
    "linspace": {f16, i32, i64},
    "linspace.tensor_overload": {f16, f32, f64, i32, i64},
    "logspace": {f16, i32, i64},
    "logspace.tensor_overload": {f16, f32, f64, i32, i64},
    "masked_logsumexp": {i64},
    "max_pool2d_with_indices_backward": {f16, f32, f64},
    "new_empty_strided": {f16},
    "nn.functional.adaptive_avg_pool3d": {f16},
    "nn.functional.adaptive_max_pool1d": {f16, f32},
    "nn.functional.adaptive_max_pool2d": {f16, f32},
    "nn.functional.bilinear": {f16},
    "nn.functional.conv_transpose1d": {f16},
    "nn.functional.conv_transpose2d": {f16},
    "nn.functional.conv_transpose3d": {f16},
    "nn.functional.cosine_similarity": {f16},
    "nn.functional.cross_entropy": {f16, f32, f64},
    "nn.functional.gaussian_nll_loss": {f16},
    "nn.functional.grid_sample": {f16, f32, f64},
    "nn.functional.interpolate.area": {f16},
    "nn.functional.nll_loss": {f16, f32, f64},
    "normal": {f16, f32, f64},
    "put": {f16, f32, f64},
    "take": {b8, f16, f32, f64, i32, i64},
    "__rdiv__": {f16},
    "__rmod__": {f16, i64},
    "__rmul__": {f16},
    "__rpow__": {f16},
    "_unsafe_masked_index": {f16},
    "_unsafe_masked_index_put_accumulate": {f16},
    "addcdiv": {f16},
    "addcmul": {f16},
    "atan2": {f16},
    "cumsum": {f16},
    "cumulative_trapezoid": {f16},
    "dist": {f16},
    "div.no_rounding_mode": {f16},
    "fmod": {f16},
    "grid_sampler_2d": {f16},
    "index_fill": {f16, f32, f64},
    "ldexp": {f16},
    "lerp": {f16},
    "linalg.householder_product": {f32},
    "linalg.matrix_norm": {f16},
    "linalg.vector_norm": {f16},
    "masked.cumsum": {f16},
    "masked.logsumexp": {f16},
    "masked.mean": {b8},
    "masked.normalize": {f16},
    "masked.prod": {f16},
    "masked.std": {f16},
    "masked.var": {f16},
    "mul": {f16},
    "nn.functional.alpha_dropout": {f16, f32, f64},
    "nn.functional.avg_pool1d": {f16, f32, f64},
    "nn.functional.avg_pool2d": {f16, f32, f64},
    "nn.functional.avg_pool3d": {f16, f32, f64},
    "nn.functional.binary_cross_entropy": {f16},
    "nn.functional.binary_cross_entropy_with_logits": {f16},
    "nn.functional.conv2d": {f16},
    "nn.functional.cosine_embedding_loss": {f16},
    "nn.functional.dropout2d": {f16, f32, f64},
    "nn.functional.dropout3d": {f16, f32, f64},
    "nn.functional.dropout": {f16, f32, f64},
    "nn.functional.feature_alpha_dropout.with_train": {f16, f32, f64},
    "nn.functional.fractional_max_pool2d": {f16, f32, f64},
    "nn.functional.fractional_max_pool3d": {f16, f32, f64},
    "nn.functional.group_norm": {f16},
    "nn.functional.hinge_embedding_loss": {f16},
    "nn.functional.interpolate.bicubic": {f16},
    "nn.functional.interpolate.bilinear": {f16},
    "nn.functional.interpolate.trilinear": {f16},
    "nn.functional.kl_div": {f16},
    "nn.functional.margin_ranking_loss": {f16},
    "nn.functional.max_pool1d": {f16, f32, f64},
    "nn.functional.max_pool3d": {f16},
    "nn.functional.mse_loss": {f16},
    "nn.functional.multi_margin_loss": {f16},
    "nn.functional.multilabel_margin_loss": {f16},
    "nn.functional.multilabel_soft_margin_loss": {f16},
    "nn.functional.normalize": {f16},
    "nn.functional.pad.replicate": {f16, f32, f64},
    "nn.functional.pad.reflect": {f16},
    "nn.functional.pairwise_distance": {f16},
    "nn.functional.poisson_nll_loss": {f16},
    "nn.functional.rms_norm": {f16},
    "norm": {f16},
    "pow": {f16},
    "prod": {f16},
    "scatter_reduce.amax": {f16, f32, f64},
    "scatter_reduce.amin": {f16, f32, f64},
    "scatter_reduce.mean": {f16, f32, f64},
    "special.xlog1py": {f16},
    "std": {f16},
    "std_mean": {f16},
    "svd_lowrank": {f32, f64},
    "trapezoid": {f16},
    "trapz": {f16},
    "true_divide": {f16},
    "var": {f16},
    "var_mean": {f16},
    "xlogy": {f16},
}

inductor_one_sample["xpu"] = {
    "_segment_reduce.lengths": {f16},
    "_segment_reduce.offsets": {f16},
    "addmv": {f16},
    "as_strided.partial_views": {f16},
    "corrcoef": {f16},
    "diff": {f16},
    "einsum": {f16, i32},
    "gradient": {f16},
    "histogram": {f32, f64},
    "histogramdd": {f32, f64},
    "index_put": {f16, f32, f64},
    "linalg.eig": {f32, f64},
    "linspace": {f16, i32, i64},
    "linspace.tensor_overload": {f16, f32, f64, i32, i64},
    "logspace": {f16, i32, i64},
    "logspace.tensor_overload": {f16, f32, f64, i32, i64},
    "masked_logsumexp": {i64},
    "max_pool2d_with_indices_backward": {f16, f32, f64},
    "new_empty_strided": {f16},
    "nn.functional.adaptive_avg_pool3d": {f16},
    "nn.functional.adaptive_max_pool1d": {f16, f32},
    "nn.functional.adaptive_max_pool2d": {f16, f32},
    "nn.functional.max_pool2d": {f16, f32, f64},
    "nn.functional.bilinear": {f16},
    "nn.functional.conv_transpose1d": {f16},
    "nn.functional.conv_transpose2d": {f16},
    "nn.functional.conv_transpose3d": {f16},
    "nn.functional.cosine_similarity": {f16},
    "nn.functional.cross_entropy": {f16, f32, f64},
    "nn.functional.gaussian_nll_loss": {f16},
    "nn.functional.grid_sample": {f16, f32, f64},
    "nn.functional.interpolate.area": {f16},
    "nn.functional.nll_loss": {f16, f32, f64},
    "normal": {f16, f32, f64},
    "put": {f16, f32, f64},
    "take": {b8, f16, f32, f64, i32, i64},
    "__rdiv__": {f16},
    "__rmod__": {f16, i64},
    "__rmul__": {f16},
    "__rpow__": {f16},
    "_unsafe_masked_index": {f16},
    "_unsafe_masked_index_put_accumulate": {f16},
    "addcdiv": {f16},
    "addcmul": {f16},
    "atan2": {f16},
    "cumsum": {f16},
    "cumulative_trapezoid": {f16},
    "dist": {f16},
    "div.no_rounding_mode": {f16},
    "fmod": {f16},
    "grid_sampler_2d": {f16},
    "index_fill": {f16, f32, f64},
    "ldexp": {f16},
    "lerp": {f16},
    "linalg.householder_product": {f32},
    "linalg.matrix_norm": {f16},
    "linalg.vector_norm": {f16},
    "masked.cumsum": {f16},
    "masked.logsumexp": {f16},
    "masked.mean": {b8},
    "masked.normalize": {f16},
    "masked.prod": {f16},
    "masked.std": {f16},
    "masked.var": {f16},
    "mul": {f16},
    "nn.functional.alpha_dropout": {f16, f32, f64},
    "nn.functional.avg_pool1d": {f16, f32, f64},
    "nn.functional.avg_pool2d": {f16, f32, f64},
    "nn.functional.avg_pool3d": {f16, f32, f64},
    "nn.functional.binary_cross_entropy": {f16},
    "nn.functional.binary_cross_entropy_with_logits": {f16},
    "nn.functional.conv2d": {f16},
    "nn.functional.cosine_embedding_loss": {f16},
    "nn.functional.dropout2d": {f16, f32, f64},
    "nn.functional.dropout3d": {f16, f32, f64},
    "nn.functional.dropout": {f16, f32, f64},
    "nn.functional.feature_alpha_dropout.with_train": {f16, f32, f64},
    "nn.functional.fractional_max_pool2d": {f16, f32, f64},
    "nn.functional.fractional_max_pool3d": {f16, f32, f64},
    "nn.functional.group_norm": {f16},
    "nn.functional.hinge_embedding_loss": {f16},
    "nn.functional.interpolate.bicubic": {f16},
    "nn.functional.interpolate.bilinear": {f16},
    "nn.functional.interpolate.trilinear": {f16},
    "nn.functional.kl_div": {f16},
    "nn.functional.margin_ranking_loss": {f16},
    "nn.functional.max_pool1d": {f16, f32, f64},
    "nn.functional.max_pool3d": {f16},
    "nn.functional.mse_loss": {f16},
    "nn.functional.multi_margin_loss": {f16},
    "nn.functional.multilabel_margin_loss": {f16},
    "nn.functional.multilabel_soft_margin_loss": {f16},
    "nn.functional.normalize": {f16},
    "nn.functional.pad.replicate": {f16, f32, f64},
    "nn.functional.pad.reflect": {f16},
    "nn.functional.pairwise_distance": {f16},
    "nn.functional.poisson_nll_loss": {f16},
    "nn.functional.rms_norm": {f16},
    "norm": {f16},
    "pow": {f16},
    "prod": {f16},
    "scatter_reduce.amax": {f16, f32, f64},
    "scatter_reduce.amin": {f16, f32, f64},
    "scatter_reduce.mean": {f16, f32, f64},
    "special.xlog1py": {f16},
    "std": {f16},
    "std_mean": {f16},
    "svd_lowrank": {f32, f64},
    "trapezoid": {f16},
    "trapz": {f16},
    "true_divide": {f16},
    "var": {f16},
    "var_mean": {f16},
    "xlogy": {f16},
}

# TODO: Fix these so strides match.
inductor_skip_exact_stride = {
    "complex",
    "empty_permuted",
    "fft.irfftn",
    "fft.irfft2",
    "linalg.diagonal",
    "linalg.eigvals",  # Fails for ROCM
    "linalg.lu",
    "linalg.lu_factor",
    "linalg.lu_factor_ex",
    "linalg.matrix_norm",
    "linalg.norm",
    "linalg.norm.subgradients_at_zero",
    "linalg.pinv.singular",
    "linalg.svdvals",
    "linalg.solve",
    "linalg.solve_ex",
    "linalg.qr",
    "lu",
    "matmul",
    "__rmatmul__",
    "nn.functional.adaptive_avg_pool1d",
    "nn.functional.group_norm",
    "nn.functional.linear",
    "nn.functional.max_pool2d",
    "nn.functional.unfold",
    "ormqr",
    "pca_lowrank",
    "rot90",
    "sum",
    "tensordot",
}

# On XPU, Inductor may apply additional layout optimizations that can change
# tensor strides compared to eager mode, so exact stride checks are relaxed
# for certain ops.
inductor_skip_exact_stride_xpu = {
    "nn.functional.conv2d",
    "nn.functional.conv_transpose2d",
    "nn.functional.max_unpool2d",
    "nn.functional.max_unpool2d.grad",
    "nn.functional.rms_norm",
}

# Custom replacements for assertEquals, in cases where a difference in value
# may not indicate correctness.


def get_sort_argsort_assert_equal_fn(is_argsort, args, kwargs):
    # Use the normal assert_equal_fn suffices for a stable sort
    if "stable" in kwargs:
        return True

    # In other cases, we need only check that the sort/argsort outputs are
    # compatible.
    orig_input = args[0]

    # The sort dimension is specified as a kwarg, or the last dimension.
    if "dim" not in kwargs:
        dim = orig_input.dim() - 1
    else:
        dim = kwargs["dim"]

    def argsort_sort_assert_equal(
        test_case_inst,
        x,
        y,
        *,
        atol=None,
        rtol=None,
        equal_nan=True,
        exact_dtype=True,
        exact_stride=False,
    ):
        if is_argsort:
            if not isinstance(x, torch.Tensor):
                raise AssertionError(f"Expected torch.Tensor, got {type(x)}")
            if not isinstance(y, torch.Tensor):
                raise AssertionError(f"Expected torch.Tensor, got {type(y)}")
        else:
            # The first tensor is the sorted values and can be asserted via
            # the usual means
            for t in (x, y):
                if not isinstance(t, tuple):
                    raise AssertionError(f"Expected tuple, got {type(t)}")
                if len(t) != 2:
                    raise AssertionError(f"Expected tuple of length 2, got {len(t)}")

            test_case_inst.assertEqual(
                x[0],
                y[0],
                atol=atol,
                rtol=rtol,
                equal_nan=equal_nan,
                exact_dtype=exact_dtype,
                exact_stride=exact_stride,
            )

            # The second tensor is the same result as an argsort.
            x = x[1]
            y = y[1]

        if exact_dtype and (x.dtype != y.dtype):
            raise AssertionError(f"The dtypes do not match: {x.dtype} != {y.dtype}.")

        if x.shape != y.shape:
            raise AssertionError(f"Shape mismatch: {x.shape} != {y.shape}")

        if exact_stride and (x.stride() != y.stride()):
            raise AssertionError(
                f"The strides do not match: {x.stride()} != {y.stride()}."
            )

        def el_to_indices(el):
            """Turn an element number into a list of indices"""
            indices = [None] * x.dim()
            for cur_dim in reversed(range(x.dim())):
                indices[cur_dim] = el % x.shape[cur_dim]
                el //= x.shape[cur_dim]
            if None in indices:
                raise AssertionError("indices contains None")
            return indices

        def get_val_by_ids(t, ids):
            """Return a value from a tensor at a given list of indices"""
            for idx in ids:
                t = t[idx]
            return t.item()

        # Loop through every value of the tensors and check for equality or
        # compatibility.
        for current_el in range(x.numel()):
            ids = el_to_indices(current_el)

            # Simple case: check equality of arsort indices
            if get_val_by_ids(x, ids) == get_val_by_ids(y, ids):
                continue

            # Complex case: check if indices refer to same value
            x_orig_ids = ids.copy()
            y_orig_ids = ids.copy()

            x_orig_ids[dim] = get_val_by_ids(x, ids)
            y_orig_ids[dim] = get_val_by_ids(y, ids)

            x_value = get_val_by_ids(orig_input, x_orig_ids)
            y_value = get_val_by_ids(orig_input, y_orig_ids)
            if x_value == y_value:
                continue

            if equal_nan:
                if math.isnan(x_value) and math.isnan(y_value):
                    continue

            raise AssertionError(
                f"Non-stable argsort outputs are incompatible at {ids}"
            )

    return argsort_sort_assert_equal


def get_argsort_assert_equal_fn(args, kwargs):
    return get_sort_argsort_assert_equal_fn(True, args, kwargs)


def get_sort_assert_equal_fn(args, kwargs):
    return get_sort_argsort_assert_equal_fn(False, args, kwargs)


CUSTOM_ASSERT_EQUALS_FNS = {
    "argsort": get_argsort_assert_equal_fn,
    "sort": get_sort_assert_equal_fn,
}


def collection_decorator(fn):
    @functools.wraps(fn)
    def inner(self, device, dtype, op):
        try:
            fn(self, device, dtype, op)
        except Exception as e:
            if COLLECT_EXPECT:
                variant = op.variant_test_name
                op_key = op.name if not variant else (op.name, variant)
                device_type = torch.device(device).type
                # failed_reasons[device_type, op_key].add(repr(e))
                seen_failed[device_type, op_key].add(dtype)
            raise e

    return inner


def _inductor_extra_samples(op_name, device, dtype, requires_grad):
    """Extra sample inputs for inductor-specific coverage.

    These exercise dynamo decomposition paths (e.g. tensor value/alpha triggering
    fma) that the shared opinfo samples don't cover.
    """
    from torch.testing._internal.common_methods_invocations import (
        make_tensor,
        S,
        SampleInput,
    )

    make_arg = partial(
        make_tensor, device=device, dtype=dtype, requires_grad=requires_grad
    )

    if op_name in ("addcmul", "addcdiv") and (
        dtype.is_floating_point or dtype.is_complex
    ):
        # Tensor value
        args = tuple(
            make_arg(shape, exclude_zero=True) for shape in ((S, S), (S, S), (S, S))
        )
        tensor_value = make_arg((), requires_grad=False)
        return [SampleInput(*args, value=tensor_value)]

    if op_name == "add" and (dtype.is_floating_point or dtype.is_complex):
        # Tensor alpha
        lhs = make_arg((S, S))
        rhs = make_arg((S, S))
        tensor_alpha = make_arg((), requires_grad=False)
        return [SampleInput(lhs, args=(rhs,), kwargs={"alpha": tensor_alpha})]

    return []


@wrapper_noop_set_seed_decorator
class TestInductorOpInfo(TestCase):
    def tearDown(self):
        torch._dynamo.reset()

    check_model = check_model
    check_model_gpu = check_model_gpu

    @onlyNativeDeviceTypes
    @suppress_warnings
    @skipCUDAMemoryLeakCheckIf(
        True
    )  # inductor kernels failing this test intermittently
    @requires_gpu_and_triton
    @skipXPUIf(
        not HAS_XPU_AND_TRITON, "Skipped! Supported XPU compiler and Triton not found"
    )
    @skipCPUIf(not HAS_CPU, "Skipped! Supported CPU compiler not found")
    @unittest.skipIf(TEST_WITH_ASAN, "Skipped under ASAN")
    @skipIfTorchDynamo("Test uses dynamo already")
    @skipIfCrossRef
    @_ops(op_db[START:END])
    @skipOps("TestInductorOpInfo", "test_comprehensive", test_skips_or_fails)
    @patch("torch._dynamo.config.raise_on_unsafe_aot_autograd", True)
    @torch._inductor.config.patch(
        {"implicit_fallbacks": False, "triton.autotune_pointwise": False}
    )
    @torch._inductor.config.patch("test_configs.runtime_triton_dtype_assert", True)
    @torch._inductor.config.patch("test_configs.static_cpp_dtype_assert", True)
    @torch._inductor.config.patch("shape_padding", False)
    @collection_decorator
    def test_comprehensive(self, device, dtype, op):
        device_type = torch.device(device).type

        if device_type not in (GPU_TYPE, "cpu"):
            raise AssertionError(f"Unexpected device_type: {device_type}")

        torch._dynamo.reset()
        with torch.no_grad():
            # TODO: should we move empty_cache to the common device interface
            if device_type == "cuda":
                torch.cuda.empty_cache()
            elif device == "xpu":
                torch.xpu.empty_cache()
        op_name = op.name
        if op.variant_test_name:
            op_name += f".{op.variant_test_name}"

        # Skip dtype=torch.uint8 for all ops except upsample and interpolate:
        allowed_dtypes = [f16, f32, f64, i32, i64, b8]
        if op_name not in (
            "nn.functional.interpolate.bilinear",
            "nn.functional.interpolate.bicubic",
            "nn.functional.upsample_bilinear",
            "nn.functional.upsample_nearest",
            "fill",
            "full_like",
        ):
            if dtype not in allowed_dtypes:
                raise unittest.SkipTest("Skipped!")

        # with open("test_output.txt", "a") as f:
        #     print(f"CONSIDERING OP {op_name} on {device_type} with {dtype} |
        # {inductor_skips[device_type].get(op_name, set())}", flush=True, file=f)
        #     print(f"CONSIDERING OP {op_name} on {device_type} with {dtype} |
        # {inductor_skips[device_type].get(op_name, set())}", flush=True)
        if dtype in inductor_skips[device_type].get(op_name, set()):
            test_expect = ExpectedTestResult.SKIP  # noqa: F841
            # with open("test_output.txt", "a") as f:
            #     print(f"SKIPPING OP {op_name} on {device_type}", flush=True, file=f)
            #     print(f"SKIPPING OP {op_name} on {device_type}", flush=True)
        elif (
            device_type == "cpu"
            and IS_LINUX
            and dtype
            in inductor_expected_failures_single_sample[device_type].get(op_name, set())
        ) or dtype in inductor_gradient_expected_failures_single_sample[
            device_type
        ].get(op_name, set()):
            test_expect = ExpectedTestResult.XFAILURE  # noqa: F841
        else:
            test_expect = ExpectedTestResult.SUCCESS  # noqa: F841

        overridden_kwargs = {}
        overridden_kwargs.update(
            inductor_override_kwargs.get(device_type, {}).get(op_name, {})
        )
        overridden_kwargs.update(
            inductor_override_kwargs.get(device_type, {}).get((op_name, dtype), {})
        )
        func = op.get_op()

        def fn(*args, **kwargs):
            return func(*args, **kwargs)

        requires_grad = (
            op.supports_autograd
            and dtype in op.supported_backward_dtypes(device_type)
            # TODO: OpInfo really ought to error out for this case, but it's
            # not exercised in test_ops_gradients atm.  The problem is not
            # complex32 per-se (which is supported by data movement only ops)
            # but that when we do backwards we expect other ops like add to work
            and dtype != torch.complex32
        )
        samples = op.sample_inputs(device, dtype, requires_grad=requires_grad)
        extra = _inductor_extra_samples(op_name, device, dtype, requires_grad)
        if extra:
            samples = itertools.chain(samples, extra)

        if (
            dtype in inductor_one_sample.get(device_type, {}).get(op_name, {})
        ) and not ALL_SAMPLES:
            if isinstance(samples, (list, tuple)):
                samples = [samples[0]]
            else:
                samples = [next(samples)]

        class HasRngOp(TorchDispatchMode):
            def __init__(self) -> None:
                super().__init__()
                self.has_rng_op = False

            def __torch_dispatch__(self, func, types, args, kwargs=None):
                kwargs = kwargs if kwargs else {}
                if torch.Tag.nondeterministic_seeded in func.tags:
                    self.has_rng_op = True

                return func(*args, **kwargs)

        def do_nopython_and_has_rng(fn, args, kwargs):
            try:
                mode = FakeTensorMode()

                def map_to_fake(e):
                    if isinstance(e, torch.Tensor):
                        return mode.from_tensor(e)
                    else:
                        return e

                args, kwargs = tree_map(map_to_fake, (args, kwargs))
                with HasRngOp() as rng_mode, mode:
                    with enable_python_dispatcher():
                        fn(*args, **kwargs)

            except (DataDependentOutputException, DynamicOutputShapeException):
                return False, rng_mode.has_rng_op

            return True, rng_mode.has_rng_op

        def get_contexts(has_rng_op, args, kwargs):
            if has_rng_op:
                # TODO - enable this, running into errors
                return (
                    # (
                    #     lambda: torch._inductor.config.patch(
                    #         {"fallback_random": True, "implicit_fallbacks": True}
                    #     ),
                    #     {"assert_equal": True},
                    # ),
                    (
                        contextlib.nullcontext,
                        {"assert_equal": False},
                    ),
                )

            ctx = functools.partial(maybe_skip_size_asserts, op)
            if op_name in CUSTOM_ASSERT_EQUALS_FNS:
                assert_equal_fn = CUSTOM_ASSERT_EQUALS_FNS[op_name](args, kwargs)
                return (
                    (
                        ctx,
                        {"assert_equal": assert_equal_fn},
                    ),
                )

            return ((ctx, {}),)

        try:

            def _get_tolerances(dtype):
                _custom_tolerances = {
                    torch.float32: (1.3e-5, 1.5e-5),
                }
                # When we are running opportunistic_fastatomics, we will expect some floating point rounding
                # errors as the order of operation is not guaranteed.
                if dtype in _custom_tolerances:
                    return _custom_tolerances[dtype]
                else:
                    return None, None

            for sample_input in samples:
                args = [sample_input.input] + list(sample_input.args)
                kwargs = sample_input.kwargs
                # UNCOMMENT TO DEBUG SEGFAULTS

                # with open("test_output.txt", "a") as f:
                #     print(f"RUNNING OP {op_name} on {device_type} with {dtype}", flush=True, file=f)
                #     print(f"RUNNING OP {op_name} on {device_type} with {dtype}", flush=True)
                rtol, atol = _get_tolerances(dtype)
                no_python, has_rng_op = do_nopython_and_has_rng(fn, args, kwargs)
                for context_fn, kwarg_overrides in get_contexts(
                    has_rng_op, args, kwargs
                ):
                    with context_fn():
                        # Base kwargs
                        adjusted_kwargs = {
                            "check_lowp": False,
                            "nopython": no_python,
                            "check_has_compiled": no_python,
                            "atol": atol,
                            "rtol": rtol,
                        }

                        # Backend-specific adjustments
                        # Triton
                        if has_triton():
                            adjusted_kwargs.update(
                                copy_to_gpu=False,
                            )
                            if device_type == GPU_TYPE:
                                adjusted_kwargs["reference_in_float"] = False

                        # skip checking gradient on CPU for now
                        if device_type == GPU_TYPE:
                            adjusted_kwargs.update(
                                check_gradient=requires_grad,
                                output_process_fn_grad=sample_input.output_process_fn_grad,
                            )
                        else:
                            adjusted_kwargs["check_gradient"] = False

                        # Update with overridden kwargs and context-specific overrides
                        adjusted_kwargs.update(overridden_kwargs)
                        adjusted_kwargs.update(kwarg_overrides)

                        # Call the appropriate check method based on device type
                        exact_stride = op_name not in inductor_skip_exact_stride
                        # XPU has additional layout optimizations that change strides differently from eager mode.
                        if exact_stride and GPU_TYPE == "xpu":
                            exact_stride = op_name not in inductor_skip_exact_stride_xpu
                        if device_type == GPU_TYPE:
                            self.check_model_gpu(
                                fn,
                                args,
                                kwargs,
                                **adjusted_kwargs,
                                exact_stride=exact_stride,
                            )
                        else:
                            self.check_model(
                                fn,
                                args,
                                kwargs,
                                **adjusted_kwargs,
                                exact_stride=exact_stride,
                            )

        except Exception as e:
            known_failure = False
            if dtype in inductor_should_fail_with_exception[device_type].get(
                op_name, set()
            ):
                failure = inductor_should_fail_with_exception[device_type][op_name][
                    dtype
                ]
                if failure in str(e):
                    known_failure = True
            if not known_failure:
                raise e

        # with open("test_output.txt", "a") as f:
        #     print(f"SUCCEEDED OP {op_name} on {device_type} with {dtype}", flush=True, file=f)


instantiate_device_type_tests(TestInductorOpInfo, globals(), allow_xpu=True)

if __name__ == "__main__":
    run_tests()



import argparse
import datetime
import logging
import re
import sys
from collections import defaultdict

import torch
from torch._C import parse_schema, Tag


FORMAT = "[%(levelname)s %(asctime)s %(filename)s:%(lineno)s] %(message)s"

log = logging.getLogger("log")
log.setLevel(logging.INFO)

handler = logging.StreamHandler()
formatter = logging.Formatter(FORMAT)
handler.setFormatter(formatter)

log.addHandler(handler)
log.propagate = False  # Avoid double logging if root logger has handlers

# How to run this test locally:
# 1 Have two virtual environments (eg conda env), one without PyTorch installed (venv_nightly)
#   one with your local changes (venv_yours).
# In venv_nightly:
# 2. First ensure that Pytorch is uninstalled, but all prereqs are installed
# 3. Install torch nightly build with
#    `pip install --pre torch -f https://download.pytorch.org/whl/nightly/cpu/torch_nightly.html`
# 4. Generate original schemas with
#    `python test/forward_backward_compatibility/dump_all_function_schemas.py --filename nightly_schemas.txt`
# Now in venv_yours:
# 5. Run this test with
#    `python test/forward_backward_compatibility/check_forward_backward_compatibility.py --existing-schemas nightly_schemas.txt`

# The date specifies how long the allowlist exclusion should apply to. Note that core ATen opset
# (https://pytorch.org/docs/stable/torch.compiler_ir.html#core-aten-ir) is guaranteed to be BC, based on this policy
# (https://dev-discuss.pytorch.org/t/core-aten-opset-backward-forward-compatibility-policy/1772) and hence the
# allowlist does not apply (or the date is always arbitrarily far for core ATen ops).
#
#   - If we NEVER give BC guarantee for an operator, you can put the
#     date arbitrarily far in the future.
#   - Otherwise, pick a date that is far enough in the future that you
#     believe you can land your diff before then.
#
# Allowlist entries can be removed after the date listed on them passes.
#
# Allowlist item format:
# [
#   0: function name regex
#   1: date until which the allowlist entry is valid
#   2: (optional, default: None) function argument regex
#   3: (optional, default: False) If True, tells us that you are NOT a core ATen op
#                                 See Note [Op removal core ATen detection]
# ]
#
# NB: function name DOES NOT include overload name!
ALLOW_LIST = [
    ("c10_experimental", datetime.date(9999, 1, 1)),
    # Internal
    ("static", datetime.date(9999, 1, 1)),
    ("prim::ModuleDictIndex", datetime.date(9999, 1, 1)),
    ("prim::MKLDNNRelu6", datetime.date(9999, 1, 1)),
    ("prim::MKLDNNRelu6_", datetime.date(9999, 1, 1)),
    ("prim::is_ort", datetime.date(9999, 1, 1)),
    ("prim::Concat", datetime.date(9999, 1, 1)),
    ("aten::_NestedTensor_GeneralizedBMM", datetime.date(9999, 1, 1)),
    # Internal, profiler-specific ops
    ("profiler::_call_end_callbacks_on_jit_fut*", datetime.date(9999, 1, 1)),
    ("profiler::_record_function_enter", datetime.date(9999, 1, 1)),
    ("aten::_cholesky_helper", datetime.date(9999, 1, 1)),
    ("aten::_cslt_sparse_mm", datetime.date(9999, 1, 1)),
    ("aten::_lstsq_helper", datetime.date(9999, 1, 1)),
    ("aten::_syevd_helper", datetime.date(9999, 1, 1)),
    ("aten::_linalg_solve_out_helper_", datetime.date(9999, 1, 1)),
    ("aten::select_backward", datetime.date(9999, 1, 1)),
    ("aten::lstsq", datetime.date(9999, 1, 1)),
    ("aten::lstsq.X", datetime.date(9999, 1, 1)),
    ("aten::slice_backward", datetime.date(9999, 1, 1)),
    ("aten::diagonal_backward", datetime.date(9999, 1, 1)),
    ("aten::rowwise_prune", datetime.date(9999, 1, 1)),
    ("aten::eig", datetime.date(9999, 1, 1)),
    ("aten::eig.e", datetime.date(9999, 1, 1)),
    ("aten::adaptive_avg_pool3d_backward", datetime.date(9999, 1, 1)),
    ("aten::_embedding_bag_dense_backward", datetime.date(9999, 1, 1)),
    ("aten::matrix_rank", datetime.date(9999, 1, 1)),
    ("aten::matrix_rank.tol", datetime.date(9999, 1, 1)),
    ("aten::randperm", datetime.date(9999, 1, 1)),
    ("aten::solve", datetime.date(9999, 1, 1)),
    ("aten::solve.solution", datetime.date(9999, 1, 1)),
    ("aten::_solve_helper", datetime.date(9999, 1, 1)),
    ("aten::_convolution_nogroup", datetime.date(9999, 1, 1)),
    ("aten::miopen_convolution_backward", datetime.date(9999, 1, 1)),
    ("aten::miopen_convolution_backward_bias", datetime.date(9999, 1, 1)),
    ("aten::miopen_convolution_backward_input", datetime.date(9999, 1, 1)),
    ("aten::miopen_convolution_backward_weight", datetime.date(9999, 1, 1)),
    ("aten::miopen_convolution_transpose_backward", datetime.date(9999, 1, 1)),
    ("aten::miopen_convolution_transpose_backward_input", datetime.date(9999, 1, 1)),
    ("aten::miopen_convolution_transpose_backward_weight", datetime.date(9999, 1, 1)),
    ("aten::miopen_depthwise_convolution_backward", datetime.date(9999, 1, 1)),
    ("aten::miopen_depthwise_convolution_backward_input", datetime.date(9999, 1, 1)),
    ("aten::miopen_depthwise_convolution_backward_weight", datetime.date(9999, 1, 1)),
    ("aten::_nested_tensor", datetime.date(9999, 1, 1)),
    ("prepacked::unpack_prepacked_sizes_conv2d", datetime.date(9999, 1, 1)),
    ("prepacked::unpack_prepacked_sizes_linear", datetime.date(9999, 1, 1)),
    ("aten::_symeig_helper", datetime.date(9999, 1, 1)),
    ("aten::symeig", datetime.date(9999, 1, 1)),
    ("aten::symeig.e", datetime.date(9999, 1, 1)),
    ("aten::native_multi_head_self_attention", datetime.date(9999, 1, 1)),
    ("aten::_native_multi_head_self_attention", datetime.date(9999, 1, 1)),
    ("aten::grid_sampler_3d_backward", datetime.date(9999, 1, 1)),
    ("aten::_transform_bias_rescale_qkv", datetime.date(9999, 1, 1)),
    ("prim::infer_squeeze_size.dim", datetime.date(9999, 1, 1)),
    ("prim::infer_squeeze_size", datetime.date(9999, 1, 1)),
    ("aten::_weight_norm_cuda_interface", datetime.date(9999, 1, 1)),
    ("aten::_weight_norm_cuda_interface_backward", datetime.date(9999, 1, 1)),
    ("aten::empty.SymInt", datetime.date(9999, 1, 1)),
    # nested tensor temporary auxiliary ops
    ("aten::_reshape_nested", datetime.date(9999, 1, 1)),
    ("aten::_reshape_nested_backward", datetime.date(9999, 1, 1)),
    ("aten::mps_linear", datetime.date(9999, 1, 1)),
    ("aten::_mps_linear", datetime.date(9999, 1, 1)),
    ("aten::_mps_max_pool2d", datetime.date(9999, 1, 1)),
    ("aten::_mps_max_pool2d.out", datetime.date(9999, 1, 1)),
    ("aten::mps_max_pool2d_backward", datetime.date(9999, 1, 1)),
    ("aten::mps_max_pool2d_backward.out", datetime.date(9999, 1, 1)),
    # TODO: FIXME: prims shouldn't be checked
    ("prims::.*", datetime.date(9999, 1, 1)),
    ("aten::_scaled_dot_product_cudnn_attention", datetime.date(9999, 1, 1)),
    # BetterTransformer 1.0 internal operators
    ("aten::_transformer_decoder_only_layer_fwd", datetime.date(9999, 1, 1)),
    ("aten::_native_decoder_only_multi_head_attention", datetime.date(9999, 1, 1)),
    # These ops were moved to python under the c10d_functional namespace
    ("aten::wait_tensor", datetime.date(9999, 1, 30)),
    ("aten::reduce_scatter_tensor", datetime.date(9999, 1, 30)),
    ("aten::all_gather_into_tensor", datetime.date(9999, 1, 30)),
    ("aten::all_reduce", datetime.date(9999, 1, 30)),
    # These ops are defined in torch/csrc/distributed/c10d/Ops.cpp
    # TODO: add back restriction when c10d ops can be exported
    ("c10d::.*", datetime.date(9999, 1, 1)),
    # Previously MPS_only did not support backward
    ("aten::_fused_rms_norm", datetime.date(2025, 12, 30)),
]

ALLOW_LIST_COMPILED = [
    (
        re.compile(item[0]),
        item[1],
        re.compile(item[2]) if (len(item) > 2 and item[2] is not None) else None,
        item[3] if len(item) > 3 else False,
    )
    for item in ALLOW_LIST
    if item[1] >= datetime.date.today()
]


def allow_listed(schema):
    for item in ALLOW_LIST_COMPILED:
        if item[0].search(str(schema)):
            if len(item) > 2 and item[2] is not None:
                # if arguments regex is present, use it
                return bool(item[2].search(str(schema))), item[3]
            return True, item[3]
    return False, None


# The nightly will fail to parse newly added syntax to schema declarations
# Add new schemas that will fail the nightly here
dont_parse_list = [
    ("_TorchScriptTesting.*", datetime.date(2099, 9, 17)),
    ("test_backend", datetime.date(2099, 9, 17)),
    ("dist_c10d", datetime.date(2099, 9, 17)),
    ("__backends__.nnc", datetime.date(2099, 9, 17)),
]


def has_valid_upgraders(schema, version_map):
    # we want to parse through the map to find if
    # the schema has valid upgraders. Since the
    # version map has entry for each overload
    # we need to do some ugly parsing.

    # the name of the operator
    schema_name = schema.name

    if schema_name not in version_map:
        return False

    entries = version_map[schema_name]

    possible_overloads = []
    possible_schemas = []
    for key, upgrader_schema_entries in entries.items():
        possible_overloads.append(key)
        possible_schemas.extend(upgrader_schema_entries)

    # let's make sure this existing schema is part of possible
    # schemas
    for old_schema in possible_schemas:
        if old_schema == schema:
            return True

    return False


def dont_parse(schema_line):
    for item in dont_parse_list:
        if item[1] < datetime.date.today():
            continue
        regexp = re.compile(item[0])
        if regexp.search(schema_line):
            return True
    return False


def load_schemas_to_dict():
    new_schemas = torch._C._jit_get_all_schemas()
    new_schemas += torch._C._jit_get_custom_class_schemas()
    new_schema_dict = defaultdict(list)
    for s in new_schemas:
        new_schema_dict[s.name].append(s)
    return new_schema_dict


def process_version_map(version_map):
    # version map maps full schema name to
    # list of upgraders. Since we only have
    # the name of the schema (aka no overload)
    # we want to first process the map to make
    # the key lookup easier. After this it will be:
    # Dict[schema_name, Dict[overload, List[schema]]]

    output = defaultdict(dict)
    for key, entries in version_map.items():
        operator_name = key.split(".")[0]
        schema_entries = [parse_schema(entry.old_schema) for entry in entries]
        output[operator_name][key] = schema_entries
    return output


def is_core_aten_op(schema) -> bool:
    # Check if the schema is a core ATen op
    if "::" not in schema.name:
        return False
    res = torch._C._get_operation_overload(schema.name, schema.overload_name)
    if res is None:
        # Note [Op removal core ATen detection]
        #
        # If the core ATen op has been removed, we cannot be sure whether it
        # was previously a core ATen op or not via checking tags this way.
        # Conservatively assume that you are ARE a core ATen op in this case.
        # This means that deleting a core ATen op will still be caught.
        # But if you're deleting an operator that is not a core ATen op
        # and add it to the allow_list, you would need to additionally specify
        # a flag in the ALLOW_LIST to tell us you are not a core ATen op.
        # See the comment block above ALLOW_LIST for more info.
        #
        # See https://github.com/pytorch/pytorch/issues/146049
        return True
    _, _, tags = res
    return Tag.core in tags


def check_bc(existing_schemas):
    new_schema_dict = load_schemas_to_dict()
    version_map = process_version_map(torch._C._get_operator_version_map())
    is_bc = True
    broken_ops = []
    for existing_schema in existing_schemas:
        is_allow_list, trust_not_core_aten = allow_listed(existing_schema)
        if is_allow_list:
            if trust_not_core_aten or not is_core_aten_op(existing_schema):
                log.info("schema: %s found on allowlist, skipping", existing_schema)
                continue
            else:
                log.info(
                    "schema: %s found on allowlist, but is a core ATen op, checking BC. "
                    "NOTE: If you have removed an operator we will conservatively assume that "
                    "it is a core ATen op. If the operator you removed is not a core ATen op, "
                    "please specify that in the ALLOW_LIST entry (see comment block on top "
                    "of ALLOW_LIST more info)",
                    existing_schema,
                )
        if has_valid_upgraders(existing_schema, version_map):
            if not is_core_aten_op(existing_schema):
                log.info("schema: %s has valid upgrader, skipping", existing_schema)
                continue
            else:
                log.info(
                    "schema: %s has a valid upgrader, but is a core ATen op, checking BC"
                )
        log.debug("processing existing schema: %s", existing_schema)
        matching_new_schemas = new_schema_dict.get(existing_schema.name, [])
        found = False
        for matching_new_schema in matching_new_schemas:
            if matching_new_schema.is_backward_compatible_with(existing_schema):
                found = True
                break
        if not found:
            log.warning(
                "Can NOT find backward compatible schemas after changes "
                "for schema %s from the following candidates:\n[\n%s\n]",
                existing_schema,
                "\n\t".join(str(s) for s in matching_new_schemas),
            )
            # TODO Print out more details about why candidates don't match.
            broken_ops.append(str(existing_schema))
            is_bc = False
    if is_bc:
        log.info("Found backward compatible schemas for all existing schemas")
    else:
        log.warning(
            "The PR is introducing backward incompatible changes to the "
            "operator library. Please contact PyTorch team to confirm "
            "whether this change is wanted or not. \n\nBroken ops: "
            "[\n\t%s\n]",
            "\n\t".join(broken_ops),
        )
    return is_bc


def check_fc(existing_schemas):
    new_schema_dict = load_schemas_to_dict()
    is_fc = True
    broken_ops = []
    for existing_schema in existing_schemas:
        is_allow_list, _ = allow_listed(existing_schema)
        if is_allow_list:
            log.info("schema: %s found on allowlist, skipping", existing_schema)
            continue
        log.info("processing existing schema: %s", existing_schema)
        matching_new_schemas = new_schema_dict.get(existing_schema.name, [])
        found = False
        possible_failure_reasons = []
        for matching_new_schema in matching_new_schemas:
            is_compatible, reason = matching_new_schema.check_forward_compatible_with(
                existing_schema
            )
            if is_compatible:
                found = True
                break
            if reason != "":
                possible_failure_reasons.append(reason)
        if not found:
            log.warning(
                "Can NOT find forward compatible schemas after changes "
                "for schema %s from the following candidates:\n[\n\t%s\n]",
                existing_schema,
                "\n\t".join(str(s) for s in matching_new_schemas),
            )
            log.warning(
                "Refer to following reasons for failure to find FC schema:\n[\n%s\n]",
                "\n\t".join(str(r) for r in possible_failure_reasons),
            )
            broken_ops.append(str(existing_schema))
            is_fc = False
    if is_fc:
        log.info("Found forward compatible schemas for all existing schemas")
    else:
        log.warning(
            "The PR is introducing a potentially forward incompatible changes to the "
            "operator library. Please contact PyTorch team to confirm "
            "whether this change is wanted or not. \n\nBroken ops: "
            "[\n\t%s\n]",
            "\n\t".join(broken_ops),
        )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Process some integers.")
    parser.add_argument(
        "--existing-schemas",
        help="filename to load existing schemas",
        type=str,
        default="schemas.txt",
    )
    args = parser.parse_args()
    existing_schema_dict = {}
    slist = []
    with open(args.existing_schemas) as f:
        while True:
            line = f.readline()
            if not line:
                break

            if dont_parse(line.strip()):
                log.info("Not parsing schema line: %s", line.strip())
                continue
            s = parse_schema(line.strip())
            slist.append(s)

    # TODO in case there is FC breaking changes,
    # we just warn for now until there is a policy.
    check_fc(slist)

    if not check_bc(slist):
        sys.exit(1)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.
# pyre-unsafe

from functools import reduce
from itertools import accumulate
from typing import List, Optional, Tuple, Type

import pytest
import torch

from mslk.attention import fmha
from mslk.attention.fmha.attn_bias import (
    BlockDiagonalPaddedKeysMask,
    PagedBlockDiagonalPaddedKeysMask,
)
from mslk.attention.fmha.common import AttentionFwOpBase
from mslk.attention.fmha.tree_attention import (
    construct_full_tree_choices,
    construct_tree_choices,
    tree_attention,
    TreeAttnMetadata,
    use_triton_splitk_for_prefix,
)
from mslk.attention.fmha.utils.bench import do_bench_cudagraph

compute_capability = (0, 0)
if torch.cuda.is_available():
    compute_capability = torch.cuda.get_device_capability("cuda")
sm80_or_better_only = pytest.mark.skipif(
    compute_capability < (8, 0), reason="requires sm80+"
)


# fmt: off
# Tree defintions - see doctring of
# xformers.ops.tree_attention.TreeAttnMetadata.from_tree_choices
# for the format description
# from https://github.com/SafeAILab/EAGLE/blob/e98fc7c/model/choices.py
eagle_mc_sim_7b_63 = [(0,),(1,),(2,),(3,),(0,0),(0,1),(0,2),(1,0),(1,1),(2,0),(2,1),(3,0)  # noqa
                ,(0,0,0),(0,0,1),(0,0,2),(0,1,0),(0,1,1),(0,2,0),(0,2,1),(1,0,0),  # noqa
                (0,0,0,0),(0,0,0,1),(0,0,0,2),(0,0,0,0,0),(0,0,0,0,1)]  # noqa

# Medusa choices from
# https://github.com/FasterDecoding/Medusa/blob/5e98053/medusa/model/medusa_choices.py
mc_sim_7b_63 = [(0,), (0, 0), (1,), (0, 1), (2,), (0, 0, 0), (1, 0), (0, 2), (3,), (0, 3), (4,), (0, 4), (2, 0), (0, 5), (0, 0, 1), (5,), (0, 6), (6,), (0, 7), (0, 1, 0), (1, 1), (7,), (0, 8), (0, 0, 2), (3, 0), (0, 9), (8,), (9,), (1, 0, 0), (0, 2, 0), (1, 2), (0, 0, 3), (4, 0), (2, 1), (0, 0, 4), (0, 0, 5), (0, 0, 0, 0), (0, 1, 1), (0, 0, 6), (0, 3, 0), (5, 0), (1, 3), (0, 0, 7), (0, 0, 8), (0, 0, 9), (6, 0), (0, 4, 0), (1, 4), (7, 0), (0, 1, 2), (2, 0, 0), (3, 1), (2, 2), (8, 0), (0, 5, 0), (1, 5), (1, 0, 1), (0, 2, 1), (9, 0), (0, 6, 0), (0, 0, 0, 1), (1, 6), (0, 7, 0)]  # noqa
vicuna_7b_stage2 = [(0,), (0, 0), (1,), (0, 1), (0, 0, 0), (1, 0), (2,), (0, 2), (0, 0, 1), (0, 3), (3,), (0, 1, 0), (2, 0), (4,), (0, 0, 2), (0, 4), (1, 1), (1, 0, 0), (0, 0, 0, 0), (5,), (0, 0, 3), (0, 5), (0, 2, 0), (3, 0), (0, 1, 1), (0, 6), (6,), (0, 7), (0, 0, 4), (4, 0), (1, 2), (0, 8), (7,), (0, 3, 0), (0, 0, 0, 1), (0, 0, 5), (2, 1), (0, 0, 6), (1, 0, 1), (0, 0, 1, 0), (2, 0, 0), (5, 0), (0, 9), (0, 1, 2), (8,), (0, 4, 0), (0, 2, 1), (1, 3), (0, 0, 7), (0, 0, 0, 2), (0, 0, 8), (1, 1, 0), (0, 1, 0, 0), (6, 0), (9,), (0, 1, 3), (0, 0, 0, 3), (1, 0, 2), (0, 5, 0), (3, 1), (0, 0, 2, 0), (7, 0), (1, 4)]  # noqa
vicuna_7b_stage1_ablation = [(0,), (0, 0), (1,), (0, 0, 0), (0, 1), (1, 0), (2,), (0, 2), (0, 0, 1), (3,), (0, 3), (0, 1, 0), (2, 0), (0, 0, 2), (0, 4), (4,), (0, 0, 0, 0), (1, 0, 0), (1, 1), (0, 0, 3), (0, 2, 0), (0, 5), (5,), (3, 0), (0, 1, 1), (0, 6), (6,), (0, 0, 4), (1, 2), (0, 0, 0, 1), (4, 0), (0, 0, 5), (0, 7), (0, 8), (0, 3, 0), (0, 0, 1, 0), (1, 0, 1), (7,), (2, 0, 0), (0, 0, 6), (2, 1), (0, 1, 2), (5, 0), (0, 2, 1), (0, 9), (0, 0, 0, 2), (0, 4, 0), (8,), (1, 3), (0, 0, 7), (0, 1, 0, 0), (1, 1, 0), (6, 0), (9,), (0, 0, 8), (0, 0, 9), (0, 5, 0), (0, 0, 2, 0), (1, 0, 2), (0, 1, 3), (0, 0, 0, 3), (3, 0, 0), (3, 1)]  # noqa
vicuna_7b_stage1 = [(0,), (0, 0), (1,), (2,), (0, 1), (1, 0), (3,), (0, 2), (4,), (0, 0, 0), (0, 3), (5,), (2, 0), (0, 4), (6,), (0, 5), (1, 1), (0, 0, 1), (7,), (3, 0), (0, 6), (8,), (9,), (0, 1, 0), (0, 7), (0, 8), (4, 0), (0, 0, 2), (1, 2), (0, 9), (2, 1), (5, 0), (1, 0, 0), (0, 0, 3), (1, 3), (0, 2, 0), (0, 1, 1), (0, 0, 4), (6, 0), (1, 4), (0, 0, 5), (2, 2), (0, 3, 0), (3, 1), (0, 0, 6), (7, 0), (1, 5), (1, 0, 1), (2, 0, 0), (0, 0, 7), (8, 0), (0, 0, 0, 0), (4, 1), (0, 1, 2), (0, 4, 0), (9, 0), (0, 2, 1), (2, 3), (1, 6), (0, 0, 8), (0, 5, 0), (3, 2), (5, 1)]  # noqa
vicuna_13b_stage2 = [(0,), (0, 0), (1,), (0, 0, 0), (0, 1), (1, 0), (2,), (0, 2), (0, 0, 1), (0, 1, 0), (3,), (0, 3), (2, 0), (0, 0, 2), (0, 0, 0, 0), (0, 4), (1, 0, 0), (1, 1), (4,), (0, 0, 3), (0, 5), (0, 2, 0), (5,), (3, 0), (0, 1, 1), (0, 6), (0, 0, 4), (0, 0, 0, 1), (0, 7), (0, 0, 5), (1, 2), (0, 0, 1, 0), (0, 3, 0), (1, 0, 1), (4, 0), (0, 0, 6), (0, 8), (2, 0, 0), (0, 9), (6,), (7,), (2, 1), (5, 0), (0, 1, 2), (0, 0, 0, 2), (8,), (0, 4, 0), (0, 1, 0, 0), (0, 2, 1), (0, 0, 7), (1, 1, 0), (1, 3), (0, 0, 2, 0), (9,), (0, 0, 8), (0, 5, 0), (0, 0, 0, 3), (0, 0, 9), (0, 1, 3), (1, 0, 2), (0, 0, 1, 1), (3, 0, 0), (1, 0, 0, 0)]  # noqa
vicuna_13b_stage1 = [(0,), (0, 0), (1,), (0, 1), (2,), (1, 0), (0, 0, 0), (0, 2), (3,), (0, 3), (4,), (2, 0), (0, 4), (0, 0, 1), (0, 5), (5,), (1, 1), (0, 1, 0), (6,), (0, 6), (0, 0, 2), (7,), (3, 0), (8,), (0, 7), (0, 8), (1, 0, 0), (0, 0, 3), (4, 0), (1, 2), (9,), (0, 9), (2, 1), (0, 2, 0), (0, 0, 4), (1, 3), (0, 1, 1), (0, 0, 5), (5, 0), (0, 3, 0), (0, 0, 0, 0), (0, 0, 6), (6, 0), (1, 4), (2, 0, 0), (0, 1, 2), (3, 1), (0, 4, 0), (1, 0, 1), (2, 2), (0, 0, 7), (1, 5), (7, 0), (0, 0, 8), (8, 0), (0, 5, 0), (0, 0, 9), (0, 2, 1), (1, 1, 0), (0, 1, 3), (4, 1), (2, 3), (1, 6)]  # noqa
vicuna_33b_stage2 = [(0,), (0, 0), (1,), (0, 1), (0, 0, 0), (1, 0), (2,), (0, 2), (0, 0, 1), (0, 3), (3,), (0, 1, 0), (2, 0), (0, 4), (4,), (0, 0, 2), (1, 1), (1, 0, 0), (0, 5), (5,), (0, 0, 0, 0), (0, 0, 3), (3, 0), (0, 2, 0), (0, 6), (0, 1, 1), (6,), (0, 0, 4), (0, 7), (7,), (1, 2), (4, 0), (8,), (0, 3, 0), (0, 0, 5), (0, 0, 0, 1), (0, 8), (2, 1), (0, 9), (1, 0, 1), (2, 0, 0), (0, 0, 6), (5, 0), (0, 0, 1, 0), (1, 3), (0, 1, 2), (0, 4, 0), (0, 0, 7), (0, 2, 1), (9,), (1, 1, 0), (0, 0, 0, 2), (6, 0), (0, 0, 8), (0, 1, 0, 0), (7, 0), (0, 1, 3), (0, 5, 0), (1, 4), (0, 0, 9), (3, 1), (1, 0, 2), (2, 2)]  # noqa
vicuna_33b_stage1 = [(0,), (1,), (0, 0), (2,), (0, 1), (3,), (1, 0), (4,), (0, 2), (5,), (0, 3), (0, 0, 0), (6,), (0, 4), (2, 0), (7,), (1, 1), (0, 5), (3, 0), (8,), (9,), (0, 6), (0, 7), (0, 0, 1), (1, 2), (4, 0), (0, 1, 0), (0, 8), (0, 9), (2, 1), (0, 0, 2), (5, 0), (1, 3), (0, 0, 3), (1, 0, 0), (1, 4), (6, 0), (0, 2, 0), (3, 1), (2, 2), (0, 0, 4), (7, 0), (0, 1, 1), (1, 5), (4, 1), (0, 0, 5), (0, 3, 0), (9, 0), (8, 0), (1, 6), (0, 0, 6), (2, 3), (0, 1, 2), (3, 2), (0, 4, 0), (2, 0, 0), (1, 7), (1, 0, 1), (0, 0, 7), (5, 1), (2, 4), (0, 0, 8), (0, 2, 1)]  # noqa
zephyr_stage2 = [(0,), (0, 0), (1,), (0, 1), (2,), (0, 0, 0), (1, 0), (0, 2), (3,), (0, 3), (4,), (2, 0), (0, 0, 1), (0, 4), (5,), (0, 5), (0, 1, 0), (1, 1), (6,), (0, 0, 2), (3, 0), (0, 6), (7,), (0, 7), (0, 8), (0, 0, 3), (1, 0, 0), (0, 9), (0, 2, 0), (1, 2), (4, 0), (8,), (9,), (2, 1), (0, 1, 1), (0, 0, 4), (0, 0, 0, 0), (5, 0), (0, 3, 0), (1, 3), (0, 0, 5), (0, 0, 6), (6, 0), (2, 0, 0), (1, 0, 1), (0, 1, 2), (0, 4, 0), (1, 4), (3, 1), (2, 2), (0, 0, 7), (7, 0), (0, 2, 1), (0, 0, 8), (0, 1, 3), (0, 5, 0), (1, 5), (0, 0, 9), (1, 1, 0), (0, 0, 0, 1), (0, 0, 1, 0), (4, 1), (2, 3)]  # noqa

medusa_choices = [
    mc_sim_7b_63,
    vicuna_7b_stage2,
    vicuna_7b_stage1_ablation,
    vicuna_7b_stage1,
    vicuna_13b_stage2,
    vicuna_13b_stage1,
    vicuna_33b_stage2,
    vicuna_33b_stage1,
    zephyr_stage2,
]

# fmt: on


@sm80_or_better_only
@pytest.mark.parametrize(
    "tree_choices",
    [
        eagle_mc_sim_7b_63,
        *medusa_choices,
        construct_full_tree_choices(1, 1),
        construct_full_tree_choices(1, 4),
        construct_full_tree_choices(4, 1),
        construct_full_tree_choices(3, 2),
        construct_full_tree_choices(2, 3),
        construct_full_tree_choices(5, 3),
        construct_full_tree_choices(10, 2),
    ],
)
@pytest.mark.parametrize("B", [1, 16])
@pytest.mark.parametrize("G", [1, 2])
@pytest.mark.parametrize("square", [True, False])
@pytest.mark.parametrize("paged", [True, False])
@torch.no_grad()
def test_tree_attention(
    tree_choices: List[Tuple[int, ...]], B: int, G: int, square: bool, paged: bool
) -> None:
    H = 8
    D = 128
    Mk = 8192
    dtype = torch.bfloat16
    tree_size_q = len(tree_choices) + 1 if square else max(len(tree_choices) // 2, 1)
    run_tree_attention_inner(
        tree_choices, B, Mk, G, H, D, tree_size_q, dtype, paged=paged
    )


class SplitKAutotune(fmha.triton_splitk.FwOp):
    AUTOTUNE = True


def run_tree_attention_inner(
    tree_choices: List[Tuple[int, ...]],
    B: int,
    Mk: int,
    G: int,
    H: int,
    D: int,
    tree_size_q: int,
    dtype: torch.dtype,
    benchmark: bool = False,
    autotune: bool = False,
    randomize_lengths: bool = True,
    paged: bool = False,
) -> None:
    """
    Test Medusa-style tree attention.
    """
    torch.manual_seed(0)
    # + 1 comes from the root node
    tree_size_kv = len(tree_choices) + 1

    # Simulate output of speculative heads
    q_full = torch.randn([B, tree_size_kv, G, H, D], device="cuda", dtype=dtype)
    q = q_full[:, -tree_size_q:].clone()
    spec_k = torch.randn([B, tree_size_kv, G, 1, D], device="cuda", dtype=dtype)
    spec_v = torch.randn_like(spec_k)
    spec_k = spec_k.expand(-1, -1, -1, H, -1)
    spec_v = spec_v.expand(-1, -1, -1, H, -1)

    # Create K/V cache before the speculative tokens
    cache_k = torch.randn([B, Mk, G, 1, D], device="cuda", dtype=dtype)
    cache_v = torch.randn_like(cache_k)
    cache_k = cache_k.expand(-1, -1, -1, H, -1)
    cache_v = cache_v.expand(-1, -1, -1, H, -1)

    if randomize_lengths:
        kv_lens_device = torch.randint(1, Mk + 1, size=(B,), device=q.device)
    else:
        kv_lens_device = torch.full(size=(B,), fill_value=Mk, device=q.device)
    kv_lens = kv_lens_device.tolist()

    triton_splitk_op = SplitKAutotune if autotune else fmha.triton_splitk.FwOp

    # Compute attention on the full context using merge_attentions: it will use
    # padded non-causal block-diagonal on the first part (before spec tokens) and
    # explicit attn_bias mask on the second part (spec tokens)
    attn = tree_attention_with_sync(
        q,
        spec_k,
        spec_v,
        cache_k,
        cache_v,
        tree_choices,
        kv_lens,
        prefix_op=triton_splitk_op,
        suffix_op=triton_splitk_op,
        paged=paged,
    )

    tree_attn_metadata = TreeAttnMetadata.from_tree_choices(
        tree_choices, q.dtype, q.device
    )

    t_optimized_ms_fa = 0.0
    t_optimized_ms_triton = 0.0
    if benchmark:
        # Construct attention bias for "left" part of the attention - the exising K/V context
        prefix_attn_bias = fmha.attn_bias.BlockDiagonalPaddedKeysMask.from_seqlens(
            q_seqlen=[tree_size_q for _ in range(B)], kv_seqlen=kv_lens, kv_padding=Mk
        )
        # Create an explit attention bias for the speculative part of the attention
        spec_attn_bias = tree_attn_metadata.attention_bias

        torch.cuda.synchronize()
        bench_stream = torch.cuda.Stream()
        with torch.cuda.stream(bench_stream):
            t_optimized_ms_triton = do_bench_cudagraph(
                lambda: tree_attention(
                    q,
                    spec_k,
                    spec_v,
                    cache_k,
                    cache_v,
                    spec_attn_bias,
                    prefix_attn_bias,
                    prefix_op=triton_splitk_op,
                    suffix_op=triton_splitk_op,
                )
            )

        torch.cuda.synchronize()
        bench_stream = torch.cuda.Stream()
        with torch.cuda.stream(bench_stream):
            t_optimized_ms_fa = do_bench_cudagraph(
                lambda: tree_attention(
                    q,
                    spec_k,
                    spec_v,
                    cache_k,
                    cache_v,
                    spec_attn_bias,
                    prefix_attn_bias,
                    prefix_op=fmha.flash.FwOp if torch.version.cuda else fmha.ck.FwOp,
                    suffix_op=triton_splitk_op,
                )
            )

    # Compute attention on the full context in a slow way, unrolling every path in the tree
    paths_w_unpadded_indices = [
        path.tolist()[:path_len]
        for path, path_len in zip(
            tree_attn_metadata.retrieval_indices, tree_attn_metadata.path_lengths
        )
    ]

    # Reference implementation, which does path unrolling, will compute attention on the full query
    # of length tree_size_kv > tree_size_q.
    # Then we'll pick only relevant elements to compare with the optimized implementation.
    paths_w_unpadded_indices_q = [
        [x - tree_size_kv + tree_size_q for x in path]
        for path in paths_w_unpadded_indices
    ]
    paths_w_unpadded_indices_q_mask = [
        x >= 0 for path in paths_w_unpadded_indices_q for x in path
    ]

    paths_w_unpadded_indices_q = [
        [x for x in path if x >= 0] for path in paths_w_unpadded_indices_q
    ]
    attn_unrolled = torch.cat(
        [attn[:, path, :, :] for path in paths_w_unpadded_indices_q], dim=1
    )
    ref_attn_full = ref_tree_attention(
        q_full, spec_k, spec_v, cache_k, cache_v, paths_w_unpadded_indices, kv_lens
    )
    ref_attn = ref_attn_full[:, paths_w_unpadded_indices_q_mask, :, :]

    if benchmark:
        # Here we compute vanilla attention with the the shapes similar to the tree attention we benchmarked above.
        # The output of this vanilla attention will be different - we are just measuring the runtime to get
        # some order of magnitude estimation of runtime.
        torch.cuda.synchronize()
        attn_bias_ref = (
            fmha.attn_bias.BlockDiagonalCausalWithOffsetPaddedKeysMask.from_seqlens(
                q_seqlen=[tree_size_q for _ in range(B)],
                kv_seqlen=kv_lens,
                kv_padding=Mk,
            )
        )
        bench_stream = torch.cuda.Stream()
        with torch.cuda.stream(bench_stream):
            t_ref_ms = do_bench_cudagraph(
                lambda: fmha.memory_efficient_attention_forward(
                    q.view(1, B * tree_size_q, G, H, D),
                    cache_k.view(1, B * Mk, G, H, D),
                    cache_v.view(1, B * Mk, G, H, D),
                    attn_bias=attn_bias_ref,
                )
            )

        gap = (1 - t_ref_ms / min(t_optimized_ms_fa, t_optimized_ms_triton)) * 100
        triton_faster = t_optimized_ms_fa > t_optimized_ms_triton
        triton_chosen = torch.version.hip is not None or use_triton_splitk_for_prefix(
            B, G, tree_size_kv
        )
        print(
            f"{B=}, {G=}, {Mk=}, {tree_size_q=}, {tree_size_kv=}: "
            f"Tree attention with FA2/CK took {t_optimized_ms_fa * 1e3:.1f}us, "
            f"with Triton Split-K took {t_optimized_ms_triton * 1e3:.1f}us, "
            f"vanilla attention took {t_ref_ms * 1e3:.1f}us, gap {gap:.2f}%. "
            f"Triton faster: {triton_faster}. "
            f"Choice optimal: {triton_chosen == triton_faster}"
        )

    torch.testing.assert_close(attn_unrolled, ref_attn, atol=1e-2, rtol=3e-3)


def ref_tree_attention(
    q: torch.Tensor,
    spec_k: torch.Tensor,
    spec_v: torch.Tensor,
    cache_k: torch.Tensor,
    cache_v: torch.Tensor,
    paths_w_unpadded_indices: List[List[int]],
    kv_lens: List[int],
) -> torch.Tensor:
    attns = []

    B, _, G, H, D = q.shape

    # Compute attention for every path separately and then concatenate

    for path in paths_w_unpadded_indices:
        q_path = q[:, path, ...]

        extra_k = torch.empty_like(spec_k[:, path, ...])
        extra_v = torch.empty_like(extra_k)

        full_k = torch.cat([cache_k, extra_k], dim=1)
        full_v = torch.cat([cache_v, extra_v], dim=1)

        tree_depth = len(path)

        for b in range(B):
            full_k[b, kv_lens[b] : kv_lens[b] + tree_depth] = spec_k[b, path]
            full_v[b, kv_lens[b] : kv_lens[b] + tree_depth] = spec_v[b, path]

        Mk = full_k.shape[1]
        attn_bias = (
            fmha.attn_bias.BlockDiagonalCausalWithOffsetPaddedKeysMask.from_seqlens(
                q_seqlen=[tree_depth for _ in range(B)],
                kv_seqlen=[s + tree_depth for s in kv_lens],
                kv_padding=Mk,
            )
        )
        attn_path = fmha.memory_efficient_attention_forward(
            q_path.view(1, B * tree_depth, G, H, D),
            full_k.view(1, B * Mk, G, H, D),
            full_v.view(1, B * Mk, G, H, D),
            attn_bias=attn_bias,
        )
        attn_path = attn_path.reshape(B, tree_depth, G, H, D)

        attns.append(attn_path)

    attn = torch.cat(attns, dim=1)

    return attn


def tree_attention_with_sync(
    q: torch.Tensor,
    spec_k: torch.Tensor,
    spec_v: torch.Tensor,
    cache_k: torch.Tensor,
    cache_v: torch.Tensor,
    tree_choices: List[Tuple[int, ...]],
    kv_lens: List[int],
    prefix_op: Optional[Type[AttentionFwOpBase]] = None,
    suffix_op: Optional[Type[AttentionFwOpBase]] = None,
    paged: bool = False,
) -> torch.Tensor:
    """
    A wrapper around tree_attention which constructs the biases.
    Arguments are the same as in xformers.ops.tree_attention.tree_attention, but instead of
    spec_attn_bias and prefix_attn_bias this function takes in tree definition in the form of tree_choices
    and K/V sequence lengths kv_lens.
    For the format of tree_choices see docstring of
    xformers.ops.tree_attention.TreeAttnMetadata.from_tree_choices.
    """
    B, tree_size_q = q.shape[:2]
    Mk = cache_k.shape[1]

    # attn_prefix ~ (B, Mk, H, D)
    prefix_attn_bias = BlockDiagonalPaddedKeysMask.from_seqlens(
        q_seqlen=[tree_size_q for _ in range(B)], kv_seqlen=kv_lens, kv_padding=Mk
    )

    if paged:
        # Create a paged K/V cache by randomly permuting blocks and storing the permutation in block_tables.
        page_size = 256
        assert Mk % page_size == 0
        max_blocks_per_row = Mk // page_size

        block_tables = torch.randperm(
            B * max_blocks_per_row, device=q.device, dtype=torch.int32
        ).view(B, max_blocks_per_row)

        cache_v = cache_v.clone()
        cache_k = cache_k.clone()

        v_view = cache_v.view(1, -1, page_size, *cache_v.shape[2:])
        k_view = cache_k.view(1, -1, page_size, *cache_k.shape[2:])

        block_tables_expanded = (
            block_tables.to(torch.int64)
            .flatten()[None, :, None, None, None, None]
            .expand(v_view.shape)
        )
        v_view.scatter_(1, block_tables_expanded, v_view.clone())
        k_view.scatter_(1, block_tables_expanded, k_view.clone())

        cache_k = k_view.view(1, -1, *cache_k.shape[2:])
        cache_v = v_view.view(1, -1, *cache_k.shape[2:])

        prefix_attn_bias = prefix_attn_bias.make_paged(  # type: ignore
            block_tables,
            page_size=page_size,
            paged_type=PagedBlockDiagonalPaddedKeysMask,
        )

    # Create an explit attention bias for the speculative part of the attention
    spec_attn_bias = TreeAttnMetadata.from_tree_choices(
        tree_choices, q.dtype, q.device
    ).attention_bias
    return tree_attention(
        q,
        spec_k,
        spec_v,
        cache_k,
        cache_v,
        spec_attn_bias[-tree_size_q:],
        prefix_attn_bias,
        prefix_op,
        suffix_op,
    )


@pytest.mark.parametrize("depth", [1, 2, 3, 4])
@pytest.mark.parametrize("branching", [1, 2, 3, 4])
def test_tree_attention_metadata_full_tree(depth: int, branching: int) -> None:
    """
    Here we test that fields tree_seq_position_ids, tree_indices, and path_lengths
    are constructed correctly; attention_bias and retrieval_indices are tested in test_tree_attention.
    """
    tree_choices = construct_full_tree_choices(depth, branching)
    tree_attn_metadata = TreeAttnMetadata.from_tree_choices(tree_choices)

    num_paths = branching**depth
    assert len(tree_attn_metadata.path_lengths) == num_paths
    assert all(length == depth + 1 for length in tree_attn_metadata.path_lengths)

    seq_pos = []
    for i in range(depth + 1):
        seq_pos.extend([i] * branching**i)

    assert seq_pos == tree_attn_metadata.tree_seq_position_ids.tolist()

    tree_indices = []
    for i in range(depth):
        tree_indices.extend(
            [i * branching + j + 1 for j in range(branching)] * branching**i
        )

    assert [0] + tree_indices == tree_attn_metadata.tree_indices.tolist()

    tree_size = len(tree_choices) + 1
    ref_child_node_indices = torch.arange(tree_size * branching).reshape(
        tree_size, branching
    )
    ref_child_node_indices[ref_child_node_indices >= tree_size - 1] = 0
    torch.testing.assert_allclose(
        tree_attn_metadata.child_node_indices, ref_child_node_indices
    )

    ref_num_children_per_node = torch.full_like(
        tree_attn_metadata.tree_indices, fill_value=branching
    )
    ref_num_children_per_node[-num_paths:] = 0
    torch.testing.assert_close(
        tree_attn_metadata.num_children_per_node, ref_num_children_per_node
    )
    ref_num_nodes_per_level = torch.tensor([branching**i for i in range(depth + 1)])
    torch.testing.assert_close(
        tree_attn_metadata.num_nodes_per_level, ref_num_nodes_per_level
    )

    ref_subtree_sizes = (
        torch.tensor([branching**i for i in range(depth + 1)]).cumsum(dim=0).tolist()
    )
    assert all(
        subtree_size == subtree_sizes_ref
        for subtree_size, subtree_sizes_ref in zip(
            tree_attn_metadata.subtree_sizes, ref_subtree_sizes
        )
    )

    ref_num_children_per_node_at_level: List[torch.Tensor] = [
        torch.tensor([branching if i < depth else 0] * (branching**i))
        for i in range(0, depth + 1)
    ]
    assert all(
        torch.allclose(num_children, ref_num_children)
        for num_children, ref_num_children in zip(
            tree_attn_metadata.num_children_per_node_at_level,
            ref_num_children_per_node_at_level,
        )
    )


@pytest.mark.parametrize("branching", [[2, 3], [2, 3, 2], [2] * 3, [4, 3, 2, 1]])
def test_tree_attention_metadata_arbitrary_tree(branching: List[int]) -> None:
    tree_choices = construct_tree_choices(branching)
    tree_attn_metadata = TreeAttnMetadata.from_tree_choices(tree_choices)
    assert all(
        length == len(branching) + 1 for length in tree_attn_metadata.path_lengths
    )

    num_paths = reduce(lambda x, y: x * y, branching)
    assert len(tree_attn_metadata.path_lengths) == num_paths

    num_nodes_per_level = [*accumulate(branching, lambda x, y: x * y)]
    seq_pos = []
    for i in range(len(num_nodes_per_level)):
        seq_pos.extend([i + 1] * num_nodes_per_level[i])
    assert [0] + seq_pos == tree_attn_metadata.tree_seq_position_ids.tolist()

    tree_indices = []
    start_idx = 1
    num_children_per_node_ref = []
    num_children_per_node_at_level_ref: List[torch.Tensor] = []
    subtree_sizes_ref: List[int] = []
    total_num_nodes = 0
    for i in range(len(branching)):
        num_nodes_prev_level = 1 if i == 0 else num_nodes_per_level[i - 1]
        tree_indices.extend(
            [start_idx + j for j in range(branching[i])] * num_nodes_prev_level
        )
        start_idx += branching[i]
        num_children_per_node_ref.extend([branching[i]] * num_nodes_prev_level)
        num_children_per_node_at_level_ref.append(
            torch.tensor([branching[i]] * num_nodes_prev_level)
        )
        total_num_nodes += num_nodes_prev_level
        subtree_sizes_ref.append(total_num_nodes)
    num_children_per_node_at_level_ref.append(
        torch.tensor([0] * num_nodes_per_level[-1])
    )
    subtree_sizes_ref.append(len(tree_indices) + 1)
    assert (
        [0] + tree_indices == tree_attn_metadata.tree_indices.tolist()
    ), f"{tree_indices=} {tree_attn_metadata.tree_indices.tolist()=}"

    num_children_per_node_ref.extend([0] * num_paths)
    num_children_per_node_ref_tensor = torch.tensor(num_children_per_node_ref)

    assert all(
        torch.allclose(num_children, num_children_ref)
        for num_children, num_children_ref in zip(
            tree_attn_metadata.num_children_per_node_at_level,
            num_children_per_node_at_level_ref,
        )
    )
    assert all(
        subtree_size == subtree_sizes_ref
        for subtree_size, subtree_sizes_ref in zip(
            tree_attn_metadata.subtree_sizes, subtree_sizes_ref
        )
    )
    torch.testing.assert_close(
        num_children_per_node_ref_tensor, tree_attn_metadata.num_children_per_node
    )
    torch.testing.assert_close(
        tree_attn_metadata.num_nodes_per_level,
        torch.tensor([1] + num_nodes_per_level),
    )
    assert tree_attn_metadata.child_node_indices.shape == (
        len(tree_indices) + 1,
        max(branching),
    )



# Owner(s): ["module: PrivateUse1"]

import _codecs
import io
import os
import tempfile
import unittest

import numpy
import torch
from torch.serialization import safe_globals
from torch.testing._internal.common_utils import (
    run_tests,
    skipIfTorchDynamo,
    TemporaryFileName,
    TestCase,
)


class TestStorage(TestCase):
    @skipIfTorchDynamo("unsupported aten.is_pinned.default")
    def test_rewrapped_storage(self):
        """Test rewrapping pinned storage"""
        pinned_a = torch.randn(10).pin_memory()
        rewrapped_a = torch.tensor((), dtype=torch.float32).set_(
            pinned_a.untyped_storage()[2:],
            size=(5,),
            stride=(1,),
            storage_offset=0,
        )
        self.assertTrue(rewrapped_a.is_pinned())
        self.assertNotEqual(pinned_a.data_ptr(), rewrapped_a.data_ptr())


class TestSerialization(TestCase):
    def test_serialization(self):
        """Test basic serialization and deserialization"""
        storage = torch.UntypedStorage(4, device=torch.device("openreg"))
        self.assertEqual(torch.serialization.location_tag(storage), "openreg:0")

        storage = torch.UntypedStorage(4, device=torch.device("openreg:0"))
        self.assertEqual(torch.serialization.location_tag(storage), "openreg:0")

        storage_cpu = torch.empty(4, 4).storage()
        storage_openreg = torch.serialization.default_restore_location(
            storage_cpu, "openreg:0"
        )
        self.assertTrue(storage_openreg.is_openreg)

        tensor = torch.empty(3, 3, device="openreg")
        self.assertEqual(torch._utils.get_tensor_metadata(tensor), {})
        metadata = {"version_number": True, "format_number": True}
        torch._utils.set_tensor_metadata(tensor, metadata)
        self.assertEqual(torch._utils.get_tensor_metadata(tensor), metadata)

        with tempfile.TemporaryDirectory() as tmpdir:
            path = os.path.join(tmpdir, "data.pt")
            torch.save(tensor, path)

            tensor_openreg = torch.load(path)
            self.assertTrue(tensor_openreg.is_openreg)
            self.assertEqual(torch._utils.get_tensor_metadata(tensor_openreg), metadata)

            tensor_cpu = torch.load(path, map_location="cpu")
            self.assertFalse(tensor_cpu.is_openreg)
            self.assertEqual(torch._utils.get_tensor_metadata(tensor_cpu), {})

    @skipIfTorchDynamo()
    @unittest.skipIf(
        numpy.__version__ < "1.25",
        "versions < 1.25 serialize dtypes differently from how it's serialized in data_legacy_numpy",
    )
    def test_open_device_numpy_serialization(self):
        """
        This tests the legacy _rebuild_device_tensor_from_numpy serialization path
        """
        data_legacy_numpy = (
            b"PK\x03\x04\x00\x00\x08\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x10\x00\x12\x00archive/data.pklFB\x0e\x00ZZZZZZZZZZZZZZ\x80\x02}q\x00X\x01"
            b"\x00\x00\x00xq\x01ctorch._utils\n_rebuild_device_tensor_from_numpy\nq\x02(cnumpy.core.m"
            b"ultiarray\n_reconstruct\nq\x03cnumpy\nndarray\nq\x04K\x00\x85q\x05c_codecs\nencode\nq\x06"
            b"X\x01\x00\x00\x00bq\x07X\x06\x00\x00\x00latin1q\x08\x86q\tRq\n\x87q\x0bRq\x0c(K\x01K\x02K"
            b"\x03\x86q\rcnumpy\ndtype\nq\x0eX\x02\x00\x00\x00f4q\x0f\x89\x88\x87q\x10Rq\x11(K\x03X\x01"
            b"\x00\x00\x00<q\x12NNNJ\xff\xff\xff\xffJ\xff\xff\xff\xffK\x00tq\x13b\x89h\x06X\x1c\x00\x00"
            b"\x00\x00\x00\xc2\x80?\x00\x00\x00@\x00\x00@@\x00\x00\xc2\x80@\x00\x00\xc2\xa0@\x00\x00\xc3"
            b"\x80@q\x14h\x08\x86q\x15Rq\x16tq\x17bctorch\nfloat32\nq\x18X\t\x00\x00\x00openreg:0q\x19\x89"
            b"tq\x1aRq\x1bs.PK\x07\x08\xdfE\xd6\xcaS\x01\x00\x00S\x01\x00\x00PK\x03\x04\x00\x00\x08"
            b"\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x11\x00.\x00"
            b"archive/byteorderFB*\x00ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZlittlePK\x07\x08"
            b"\x85=\xe3\x19\x06\x00\x00\x00\x06\x00\x00\x00PK\x03\x04\x00\x00\x08\x08\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x0f\x00=\x00archive/versionFB9\x00"
            b"ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ3\nPK\x07\x08\xd1\x9egU\x02\x00\x00"
            b"\x00\x02\x00\x00\x00PK\x03\x04\x00\x00\x08\x08\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x1e\x002\x00archive/.data/serialization_idFB.\x00ZZZZZZZZZZZZZ"
            b"ZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZZ0636457737946401051300000025273995036293PK\x07\x08\xee(\xcd"
            b"\x8d(\x00\x00\x00(\x00\x00\x00PK\x01\x02\x00\x00\x00\x00\x08\x08\x00\x00\x00\x00\x00\x00"
            b"\xdfE\xd6\xcaS\x01\x00\x00S\x01\x00\x00\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00archive/data.pklPK\x01\x02\x00\x00\x00\x00\x08\x08\x00\x00\x00\x00"
            b"\x00\x00\x85=\xe3\x19\x06\x00\x00\x00\x06\x00\x00\x00\x11\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\xa3\x01\x00\x00archive/byteorderPK\x01\x02\x00\x00\x00\x00\x08\x08\x00"
            b"\x00\x00\x00\x00\x00\xd1\x9egU\x02\x00\x00\x00\x02\x00\x00\x00\x0f\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x16\x02\x00\x00archive/versionPK\x01\x02\x00\x00\x00\x00\x08"
            b"\x08\x00\x00\x00\x00\x00\x00\xee(\xcd\x8d(\x00\x00\x00(\x00\x00\x00\x1e\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x92\x02\x00\x00archive/.data/serialization_idPK\x06"
            b"\x06,\x00\x00\x00\x00\x00\x00\x00\x1e\x03-\x00\x00\x00\x00\x00\x00\x00\x00\x00\x04\x00\x00"
            b"\x00\x00\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x06\x01\x00\x00\x00\x00\x00\x008\x03\x00"
            b"\x00\x00\x00\x00\x00PK\x06\x07\x00\x00\x00\x00>\x04\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00"
            b"PK\x05\x06\x00\x00\x00\x00\x04\x00\x04\x00\x06\x01\x00\x008\x03\x00\x00\x00\x00"
        )
        buf_data_legacy_numpy = io.BytesIO(data_legacy_numpy)

        with safe_globals(
            [
                (
                    (
                        numpy.core.multiarray._reconstruct,
                        "numpy.core.multiarray._reconstruct",
                    )
                    if numpy.__version__ >= "2.1"
                    else numpy.core.multiarray._reconstruct
                ),
                numpy.ndarray,
                numpy.dtype,
                _codecs.encode,
                numpy.dtypes.Float32DType,
            ]
        ):
            sd_loaded = torch.load(buf_data_legacy_numpy, weights_only=True)
            buf_data_legacy_numpy.seek(0)
            # Test map_location
            sd_loaded_cpu = torch.load(
                buf_data_legacy_numpy, weights_only=True, map_location="cpu"
            )

        expected = torch.tensor(
            [[1, 2, 3], [4, 5, 6]], dtype=torch.float32, device="openreg"
        )
        self.assertEqual(sd_loaded["x"].cpu(), expected.cpu())
        self.assertFalse(sd_loaded["x"].is_cpu)
        self.assertTrue(sd_loaded_cpu["x"].is_cpu)

    def test_open_device_cpu_serialization(self):
        default_protocol = torch.serialization.DEFAULT_PROTOCOL

        with unittest.mock.patch.object(torch._C, "_has_storage", return_value=False):
            x = torch.randn(2, 3)
            x_openreg = x.to("openreg")
            sd = {"x": x_openreg}
            rebuild_func = x_openreg._reduce_ex_internal(default_protocol)[0]
            self.assertTrue(
                rebuild_func is torch._utils._rebuild_device_tensor_from_cpu_tensor
            )

            # Test map_location
            with TemporaryFileName() as f:
                torch.save(sd, f)
                sd_loaded = torch.load(f, weights_only=True)
                # Test map_location
                sd_loaded_cpu = torch.load(f, weights_only=True, map_location="cpu")
            self.assertFalse(sd_loaded["x"].is_cpu)
            self.assertEqual(sd_loaded["x"].cpu(), x)
            self.assertTrue(sd_loaded_cpu["x"].is_cpu)

            # Test metadata_only
            with TemporaryFileName() as f:
                with self.assertRaisesRegex(
                    RuntimeError,
                    "Cannot serialize tensors on backends with no storage under skip_data context manager",
                ):
                    with torch.serialization.skip_data():
                        torch.save(sd, f)

    def test_serialization_metadata_preservation(self):
        """Test that metadata is preserved during serialization"""
        tensor = torch.empty(3, 3, device="openreg")
        metadata = {"version_number": True, "format_number": True}
        torch._utils.set_tensor_metadata(tensor, metadata)

        with tempfile.TemporaryDirectory() as tmpdir:
            path = os.path.join(tmpdir, "data.pt")
            torch.save(tensor, path)

            loaded_tensor = torch.load(path)
            self.assertEqual(torch._utils.get_tensor_metadata(loaded_tensor), metadata)

    def test_serialization_map_location_cpu(self):
        """Test serialization with map_location to CPU"""
        tensor = torch.randn(3, 3, device="openreg")

        with tempfile.TemporaryDirectory() as tmpdir:
            path = os.path.join(tmpdir, "data.pt")
            torch.save(tensor, path)

            loaded_cpu = torch.load(path, map_location="cpu")
            self.assertTrue(loaded_cpu.is_cpu)
            self.assertEqual(loaded_cpu, tensor.cpu())

    def test_serialization_map_location_device(self):
        """Test serialization with map_location to specific device"""
        tensor = torch.randn(3, 3, device="openreg:0")

        with tempfile.TemporaryDirectory() as tmpdir:
            path = os.path.join(tmpdir, "data.pt")
            torch.save(tensor, path)

            loaded_device = torch.load(path, map_location="openreg:1")
            self.assertEqual(loaded_device.device.index, 1)
            self.assertEqual(loaded_device.cpu(), tensor.cpu())

    def test_serialization_storage_location_tag(self):
        """Test storage location tag"""
        storage = torch.UntypedStorage(4, device=torch.device("openreg:1"))
        self.assertEqual(torch.serialization.location_tag(storage), "openreg:1")

        storage = torch.UntypedStorage(4, device=torch.device("openreg"))
        self.assertEqual(torch.serialization.location_tag(storage), "openreg:0")

    def test_serialization_default_restore_location(self):
        """Test default restore location"""
        storage_cpu = torch.empty(4, 4).storage()

        storage_openreg0 = torch.serialization.default_restore_location(
            storage_cpu, "openreg:0"
        )
        self.assertTrue(storage_openreg0.is_openreg)

        storage_openreg1 = torch.serialization.default_restore_location(
            storage_cpu, "openreg:1"
        )
        self.assertTrue(storage_openreg1.is_openreg)


if __name__ == "__main__":
    run_tests()



# This backend is maintained by ONNX team. To direct issues
# to the right people, please tag related GitHub issues with `module: onnx`.
#
# Maintainers' Github IDs: wschin, xadupre
# from torch.onnx._internal.onnxruntime import (
#     is_onnxrt_backend_supported,
#     torch_compile_backend,
# )

# from .registry import register_backend

"""
Placeholder for onnxruntime backend for dynamo
"""

# def has_onnxruntime():
#     # FIXME: update test/dynamo/test_backends.py to call is_onnxrt_backend_supported()
#     return is_onnxrt_backend_supported()


# if is_onnxrt_backend_supported():
#     register_backend(name="onnxrt", compiler_fn=torch_compile_backend)
# else:

#     def information_displaying_backend(*args, **kwargs):
#         raise ImportError(
#             "onnxrt is not registered as a backend. "
#             "Please make sure all dependencies such as "
#             "numpy, onnx, onnxscript, and onnxruntime-training are installed. "
#             "Suggested procedure to fix dependency problem:\n"
#             "  (1) pip or conda install numpy onnx onnxscript onnxruntime-training.\n"
#             "  (2) Open a new python terminal.\n"
#             "  (3) Call the API `torch.onnx.is_onnxrt_backend_supported()`:\n"
#             "  (4)   If it returns `True`, then you can use `onnxrt` backend.\n"
#             "  (5)   If it returns `False`, please execute the package importing section in "
#             "torch/onnx/_internal/onnxruntime.py under pdb line-by-line to see which import fails."
#         )

#     register_backend(name="onnxrt", compiler_fn=information_displaying_backend)



# Owner(s): ["oncall: distributed checkpointing"]
import os
import shutil
import tempfile
from typing import Any

import torch
from torch.distributed.checkpoint._experimental.checkpoint_reader import (
    CheckpointReader,
)
from torch.distributed.checkpoint._experimental.types import RankInfo
from torch.testing._internal.common_utils import run_tests, TestCase


class TestCheckpointReader(TestCase):
    def setUp(self):
        super().setUp()
        # Create a temporary directory for test checkpoints
        self.temp_dir = tempfile.mkdtemp()

        # Create test objects
        self.rank_info = RankInfo(
            global_rank=0,
            global_world_size=1,
        )

        # Create the checkpoint reader
        self.reader = CheckpointReader(
            rank_info=self.rank_info,
        )

        # Create a test state dictionary
        self.state_dict = {
            "model": {
                "weight": torch.randn(10, 5),
                "bias": torch.randn(5),
                "test_list": [torch.randn(2), torch.randn(2)],
            },
            "optimizer": {
                "param_groups": [
                    {"lr": 0.01, "test_list": [torch.randn(2), torch.randn(2)]}
                ]
            },
            "epoch": 5,
            "step": 1000,
        }

        # Create a test checkpoint file
        self.checkpoint_path = os.path.join(self.temp_dir, "checkpoint")
        os.makedirs(self.checkpoint_path, exist_ok=True)
        checkpoint_file = os.path.join(
            self.checkpoint_path, f"checkpoint_{self.rank_info.global_rank}.pt"
        )
        torch.save(self.state_dict, checkpoint_file)

    def move_tensors_to_device(self, state_dict: Any, device: str) -> Any:
        """
        Recursively move all tensors in a nested dictionary to CUDA.

        Args:
            state_dict (dict): A dictionary potentially containing nested dictionaries and tensors.

        Returns:
            dict: A new dictionary with all tensors moved to CUDA.
        """
        if isinstance(state_dict, dict):
            return {
                key: self.move_tensors_to_device(value, device)
                for key, value in state_dict.items()
            }
        elif isinstance(state_dict, list):
            return [self.move_tensors_to_device(item, device) for item in state_dict]
        elif isinstance(state_dict, torch.Tensor):
            return state_dict.cuda() if device == "cpu" else state_dict.cpu()
        else:
            return state_dict

    def deep_compare(self, obj1: Any, obj2: Any) -> bool:
        if isinstance(obj1, dict) and isinstance(obj2, dict):
            if obj1.keys() != obj2.keys():
                return False
            return all(self.deep_compare(obj1[key], obj2[key]) for key in obj1)
        elif isinstance(obj1, (list, tuple)) and isinstance(obj2, (list, tuple)):
            if len(obj1) != len(obj2):
                return False
            return all(
                self.deep_compare(item1, item2) for item1, item2 in zip(obj1, obj2)
            )
        elif isinstance(obj1, torch.Tensor) and isinstance(obj2, torch.Tensor):
            return torch.equal(obj1, obj2)
        else:
            return obj1 == obj2

    def tearDown(self):
        # Clean up the temporary directory
        shutil.rmtree(self.temp_dir)

    def test_read_checkpoint(self):
        """Test that read correctly reads a checkpoint file."""

        # Call read
        read_state_dict, missing_keys = self.reader.read(self.checkpoint_path)
        self.assertEqual(missing_keys, [])

        # Verify that the read state dictionary contains the expected values
        self.assertIn("model", read_state_dict)
        self.assertIn("optimizer", read_state_dict)
        self.assertTrue(self.deep_compare(read_state_dict, self.state_dict))

        # No hooks to verify since we removed them

    def test_read_with_map_location(self):
        """Test that read correctly uses the map_location parameter."""
        # Call read with map_location='cpu'
        map_location = "cuda" if torch.cuda.is_available() else "cpu"
        read_state_dict, _ = self.reader.read(
            self.checkpoint_path, map_location=map_location
        )

        # Verify that the read state dictionary contains the expected values
        self.assertIn("model", read_state_dict)
        self.assertIn("optimizer", read_state_dict)
        self.assertEqual(read_state_dict["epoch"], 5)
        self.assertEqual(read_state_dict["step"], 1000)
        self.assertEqual(read_state_dict["model"]["weight"].device.type, map_location)
        read_state_dict, _ = self.reader.read(
            self.checkpoint_path, map_location=map_location
        )

        # Verify that the read state dictionary contains the expected values
        self.assertIn("model", read_state_dict)
        self.assertIn("optimizer", read_state_dict)
        self.assertEqual(read_state_dict["epoch"], 5)
        self.assertEqual(read_state_dict["step"], 1000)
        self.assertEqual(read_state_dict["model"]["weight"].device.type, map_location)

    def test_read_nonexistent_checkpoint(self):
        """Test that read raises FileNotFoundError for a nonexistent checkpoint."""
        # Set up a path to a nonexistent checkpoint
        nonexistent_path = os.path.join(self.temp_dir, "nonexistent_checkpoint")

        # Call read and expect a FileNotFoundError
        with self.assertRaises(FileNotFoundError):
            self.reader.read(nonexistent_path)

    def test_read_with_kwargs(self):
        """Test that read correctly passes kwargs."""
        # Call read with additional kwargs
        kwargs = {"extra": "value"}
        self.reader.read(self.checkpoint_path, **kwargs)

    def test_partial_read(self):
        """Test that read with state_dict correctly loads only the requested keys."""
        # Create a partial state dictionary with only some keys
        partial_state_dict = {}
        partial_state_dict["optimizer"] = None
        partial_state_dict["model"] = {"weight": torch.randn(10, 5)}
        partial_state_dict["epoch"] = None
        # Call read with state_dict
        updated_state_dict, _ = self.reader.read(
            self.checkpoint_path,
            partial_state_dict,
        )

        # Verify that the updated state dictionary contains values from both dictionaries
        self.assertIn("model", updated_state_dict)
        self.assertIn("epoch", updated_state_dict)
        self.assertTrue(
            torch.equal(
                updated_state_dict["model"]["weight"],
                self.state_dict["model"]["weight"],
            )
        )

        self.assertTrue(
            self.deep_compare(
                updated_state_dict["optimizer"], self.state_dict["optimizer"]
            )
        )
        self.assertEqual(updated_state_dict["epoch"], 5)  # From checkpoint

        self.assertNotIn("bias", updated_state_dict["model"])
        self.assertNotIn("step", updated_state_dict)

    def test_partial_read_missing_keys(self):
        """Test that partial_read correctly reports missing keys."""
        # Create a partial state dictionary with keys that don't exist in the checkpoint
        partial_state_dict = {
            "model": None,
            "nonexistent_key": None,  # This key doesn't exist in the checkpoint
            "another_missing_key": {"nested": None},  # This key also doesn't exist
        }

        # Call read with state_dict
        _, missing_keys = self.reader.read(
            self.checkpoint_path,
            partial_state_dict,
        )

        # Verify that missing keys are correctly reported
        self.assertIn("nonexistent_key", missing_keys)
        self.assertIn("another_missing_key", missing_keys)

        # Verify that keys that exist in the checkpoint are not in missing_keys
        self.assertNotIn("model", missing_keys)

    def test_partial_read_different_dtypes(self):
        """Test that partial_read correctly handles different tensor dtypes."""
        # Create a state dictionary with tensors of different dtypes
        dtype_state_dict = {
            "float32": torch.randn(10, 10, dtype=torch.float32),
            "float64": torch.randn(10, 10, dtype=torch.float64),
            "int32": torch.randint(-100, 100, (10, 10), dtype=torch.int32),
            "int64": torch.randint(-100, 100, (10, 10), dtype=torch.int64),
            "bool": torch.randint(0, 2, (10, 10), dtype=torch.bool),
        }

        # Save the state dictionary
        dtype_checkpoint_path = os.path.join(self.temp_dir, "dtype_checkpoint")
        os.makedirs(dtype_checkpoint_path, exist_ok=True)
        checkpoint_file = os.path.join(
            dtype_checkpoint_path, f"checkpoint_{self.rank_info.global_rank}.pt"
        )
        torch.save(dtype_state_dict, checkpoint_file)

        # Create a partial state dictionary requesting tensors of each dtype
        partial_state_dict = {
            "float32": torch.randn(10, 10, dtype=torch.float32),
            "float64": None,
            "int32": None,
            "int64": None,
            "bool": None,
        }

        # Load the partial state dictionary
        updated_state_dict, _ = self.reader.read(
            os.path.dirname(checkpoint_file),
            partial_state_dict,
        )

        # Verify that tensors of each dtype were loaded correctly
        for key in dtype_state_dict:
            self.assertIn(key, updated_state_dict)
            self.assertEqual(updated_state_dict[key].dtype, dtype_state_dict[key].dtype)
            self.assertTrue(
                torch.allclose(updated_state_dict[key], dtype_state_dict[key])
            )


if __name__ == "__main__":
    run_tests()



# Defined in torch/csrc/export/pt2_archive_constants.h
from torch._C._export import pt2_archive_constants


AOTINDUCTOR_DIR: str = pt2_archive_constants.AOTINDUCTOR_DIR
ARCHIVE_FORMAT_PATH: str = pt2_archive_constants.ARCHIVE_FORMAT_PATH
ARCHIVE_FORMAT_VALUE: str = pt2_archive_constants.ARCHIVE_FORMAT_VALUE
ARCHIVE_ROOT_NAME: str = pt2_archive_constants.ARCHIVE_ROOT_NAME
ARCHIVE_VERSION_PATH: str = pt2_archive_constants.ARCHIVE_VERSION_PATH
ARCHIVE_VERSION_VALUE: str = pt2_archive_constants.ARCHIVE_VERSION_VALUE
CONSTANTS_DIR: str = pt2_archive_constants.CONSTANTS_DIR
CONSTANTS_CONFIG_FILENAME_FORMAT: str = (
    pt2_archive_constants.CONSTANTS_CONFIG_FILENAME_FORMAT
)
CUSTOM_OBJ_FILENAME_PREFIX: str = pt2_archive_constants.CUSTOM_OBJ_FILENAME_PREFIX
EXECUTORCH_DIR: str = pt2_archive_constants.EXECUTORCH_DIR
EXTRA_DIR: str = pt2_archive_constants.EXTRA_DIR
MODELS_DIR: str = pt2_archive_constants.MODELS_DIR
MODELS_FILENAME_FORMAT: str = pt2_archive_constants.MODELS_FILENAME_FORMAT
MODULE_INFO_PATH: str = pt2_archive_constants.MODULE_INFO_PATH
MTIA_DIR: str = pt2_archive_constants.MTIA_DIR
SAMPLE_INPUTS_DIR: str = pt2_archive_constants.SAMPLE_INPUTS_DIR
SAMPLE_INPUTS_FILENAME_FORMAT: str = pt2_archive_constants.SAMPLE_INPUTS_FILENAME_FORMAT
TENSOR_CONSTANT_FILENAME_PREFIX: str = (
    pt2_archive_constants.TENSOR_CONSTANT_FILENAME_PREFIX
)
WEIGHTS_CONFIG_FILENAME_FORMAT: str = (
    pt2_archive_constants.WEIGHTS_CONFIG_FILENAME_FORMAT
)
WEIGHT_FILENAME_PREFIX: str = pt2_archive_constants.WEIGHT_FILENAME_PREFIX
WEIGHTS_DIR: str = pt2_archive_constants.WEIGHTS_DIR
XL_MODEL_WEIGHTS_DIR: str = pt2_archive_constants.XL_MODEL_WEIGHTS_DIR
XL_MODEL_WEIGHTS_PARAM_CONFIG_PATH: str = (
    pt2_archive_constants.XL_MODEL_WEIGHTS_PARAM_CONFIG_PATH
)



# mypy: allow-untyped-defs
import torch
from torch._ops import OpOverload, OpOverloadPacket


def _register_decomposition(op: OpOverload, graph: torch._C.Graph) -> None:
    if isinstance(op, OpOverloadPacket):
        raise AssertionError(
            f"Must pass specific op overload, not overload packet, found {op}"
        )
    if not isinstance(op, OpOverload):
        raise AssertionError(f"Expected OpOverload, got {type(op)}")

    torch._C._jit_register_decomposition_for_schema(op._schema, graph)
